from __future__ import annotations

import pytest

from PIL import Image

from .helper import CachedProperty, assert_image_equal


class TestImagingPaste:
    size = 128

    def assert_9points_image(
        self, im: Image.Image, expected: list[tuple[int, int, int, int]]
    ) -> None:
        px = im.load()
        assert px is not None
        actual = [
            px[0, 0],
            px[self.size // 2, 0],
            px[self.size - 1, 0],
            px[0, self.size // 2],
            px[self.size // 2, self.size // 2],
            px[self.size - 1, self.size // 2],
            px[0, self.size - 1],
            px[self.size // 2, self.size - 1],
            px[self.size - 1, self.size - 1],
        ]
        assert actual == [
            point[0] if im.mode == "L" else point[: len(im.mode)] for point in expected
        ]

    def assert_9points_paste(
        self,
        im: Image.Image,
        im2: Image.Image | str | tuple[int, ...],
        mask: Image.Image,
        expected: list[tuple[int, int, int, int]],
    ) -> None:
        im3 = im.copy()
        im3.paste(im2, (0, 0), mask)
        self.assert_9points_image(im3, expected)

        # Abbreviated syntax
        im.paste(im2, mask)
        self.assert_9points_image(im, expected)

    @CachedProperty
    def mask_1(self) -> Image.Image:
        mask = Image.new("1", (self.size, self.size))
        px = mask.load()
        assert px is not None
        for y in range(mask.height):
            for x in range(mask.width):
                px[y, x] = (x + y) % 2
        return mask

    @CachedProperty
    def mask_L(self) -> Image.Image:
        return self.gradient_L.transpose(Image.Transpose.ROTATE_270)

    @CachedProperty
    def gradient_L(self) -> Image.Image:
        gradient = Image.new("L", (self.size, self.size))
        px = gradient.load()
        assert px is not None
        for y in range(gradient.height):
            for x in range(gradient.width):
                px[y, x] = (x + y) % 255
        return gradient

    @CachedProperty
    def gradient_RGB(self) -> Image.Image:
        return Image.merge(
            "RGB",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.Transpose.ROTATE_90),
                self.gradient_L.transpose(Image.Transpose.ROTATE_180),
            ],
        )

    @CachedProperty
    def gradient_LA(self) -> Image.Image:
        return Image.merge(
            "LA",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.Transpose.ROTATE_90),
            ],
        )

    @CachedProperty
    def gradient_RGBA(self) -> Image.Image:
        return Image.merge(
            "RGBA",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.Transpose.ROTATE_90),
                self.gradient_L.transpose(Image.Transpose.ROTATE_180),
                self.gradient_L.transpose(Image.Transpose.ROTATE_270),
            ],
        )

    @CachedProperty
    def gradient_RGBa(self) -> Image.Image:
        return Image.merge(
            "RGBa",
            [
                self.gradient_L,
                self.gradient_L.transpose(Image.Transpose.ROTATE_90),
                self.gradient_L.transpose(Image.Transpose.ROTATE_180),
                self.gradient_L.transpose(Image.Transpose.ROTATE_270),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_solid(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "red")
        im2 = getattr(self, "gradient_" + mode)

        im.paste(im2, (12, 23))

        im = im.crop((12, 23, im2.width + 12, im2.height + 23))
        assert_image_equal(im, im2)

    @pytest.mark.parametrize("y", [10, -10])
    @pytest.mark.parametrize("mode", ["L", "RGB"])
    @pytest.mark.parametrize("mask_mode", ["", "1", "L", "LA", "RGBa"])
    def test_image_self(self, y: int, mode: str, mask_mode: str) -> None:
        im = getattr(self, "gradient_" + mode)
        mask = Image.new(mask_mode, im.size, 0xFFFFFFFF) if mask_mode else None

        im_self = im.copy()
        im_self.paste(im_self, (0, y), mask)

        im_copy = im.copy()
        im_copy.paste(im_copy.copy(), (0, y), mask)

        assert_image_equal(im_self, im_copy)

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_mask_1(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "white")
        im2 = getattr(self, "gradient_" + mode)

        self.assert_9points_paste(
            im,
            im2,
            self.mask_1,
            [
                (255, 255, 255, 255),
                (255, 255, 255, 255),
                (127, 254, 127, 0),
                (255, 255, 255, 255),
                (255, 255, 255, 255),
                (191, 190, 63, 64),
                (127, 0, 127, 254),
                (191, 64, 63, 190),
                (255, 255, 255, 255),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_mask_L(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "white")
        im2 = getattr(self, "gradient_" + mode)

        self.assert_9points_paste(
            im,
            im2,
            self.mask_L,
            [
                (128, 191, 255, 191),
                (208, 239, 239, 208),
                (255, 255, 255, 255),
                (112, 111, 206, 207),
                (192, 191, 191, 191),
                (239, 239, 207, 207),
                (128, 1, 128, 254),
                (207, 113, 112, 207),
                (255, 191, 128, 191),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_mask_LA(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "white")
        im2 = getattr(self, "gradient_" + mode)

        self.assert_9points_paste(
            im,
            im2,
            self.gradient_LA,
            [
                (128, 191, 255, 191),
                (112, 207, 206, 111),
                (128, 254, 128, 1),
                (208, 208, 239, 239),
                (192, 191, 191, 191),
                (207, 207, 112, 113),
                (255, 255, 255, 255),
                (239, 207, 207, 239),
                (255, 191, 128, 191),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_mask_RGBA(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "white")
        im2 = getattr(self, "gradient_" + mode)

        self.assert_9points_paste(
            im,
            im2,
            self.gradient_RGBA,
            [
                (128, 191, 255, 191),
                (208, 239, 239, 208),
                (255, 255, 255, 255),
                (112, 111, 206, 207),
                (192, 191, 191, 191),
                (239, 239, 207, 207),
                (128, 1, 128, 254),
                (207, 113, 112, 207),
                (255, 191, 128, 191),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_image_mask_RGBa(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "white")
        im2 = getattr(self, "gradient_" + mode)

        self.assert_9points_paste(
            im,
            im2,
            self.gradient_RGBa,
            [
                (128, 255, 126, 255),
                (0, 127, 126, 255),
                (126, 253, 126, 255),
                (128, 127, 254, 255),
                (0, 255, 254, 255),
                (126, 125, 254, 255),
                (128, 1, 128, 255),
                (0, 129, 128, 255),
                (126, 255, 128, 255),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_color_solid(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), "black")

        rect = (12, 23, 128 + 12, 128 + 23)
        im.paste("white", rect)

        hist = im.crop(rect).histogram()
        while hist:
            head, hist = hist[:256], hist[256:]
            assert head[255] == 128 * 128
            assert sum(head[:255]) == 0

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_color_mask_1(self, mode: str) -> None:
        im = Image.new(mode, (200, 200), (50, 60, 70, 80)[: len(mode)])
        color = (10, 20, 30, 40)[: len(mode)]

        self.assert_9points_paste(
            im,
            color,
            self.mask_1,
            [
                (50, 60, 70, 80),
                (50, 60, 70, 80),
                (10, 20, 30, 40),
                (50, 60, 70, 80),
                (50, 60, 70, 80),
                (10, 20, 30, 40),
                (10, 20, 30, 40),
                (10, 20, 30, 40),
                (50, 60, 70, 80),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_color_mask_L(self, mode: str) -> None:
        im = getattr(self, "gradient_" + mode).copy()
        color = "white"

        self.assert_9points_paste(
            im,
            color,
            self.mask_L,
            [
                (127, 191, 254, 191),
                (111, 207, 206, 110),
                (127, 254, 127, 0),
                (207, 207, 239, 239),
                (191, 191, 190, 191),
                (207, 206, 111, 112),
                (254, 254, 254, 255),
                (239, 206, 206, 238),
                (254, 191, 127, 191),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_color_mask_RGBA(self, mode: str) -> None:
        im = getattr(self, "gradient_" + mode).copy()
        color = "white"

        self.assert_9points_paste(
            im,
            color,
            self.gradient_RGBA,
            [
                (127, 191, 254, 191),
                (111, 207, 206, 110),
                (127, 254, 127, 0),
                (207, 207, 239, 239),
                (191, 191, 190, 191),
                (207, 206, 111, 112),
                (254, 254, 254, 255),
                (239, 206, 206, 238),
                (254, 191, 127, 191),
            ],
        )

    @pytest.mark.parametrize("mode", ["RGBA", "RGB", "L"])
    def test_color_mask_RGBa(self, mode: str) -> None:
        im = getattr(self, "gradient_" + mode).copy()
        color = "white"

        self.assert_9points_paste(
            im,
            color,
            self.gradient_RGBa,
            [
                (255, 63, 126, 63),
                (47, 143, 142, 46),
                (126, 253, 126, 255),
                (15, 15, 47, 47),
                (63, 63, 62, 63),
                (142, 141, 46, 47),
                (255, 255, 255, 0),
                (48, 15, 15, 47),
                (126, 63, 255, 63),
            ],
        )

    def test_different_sizes(self) -> None:
        im = Image.new("RGB", (100, 100))
        im2 = Image.new("RGB", (50, 50))

        im.copy().paste(im2)
        im.copy().paste(im2, (0, 0))

    @pytest.mark.parametrize(
        "box",
        (
            (2**31 - 1, 0, -(2**31), 1),
            (0, 2**31 - 1, 1, -(2**31)),
        ),
    )
    def test_overflow(self, box: tuple[int, int, int, int]) -> None:
        im = Image.new("1", (1, 1))
        im.paste(1, box)

        with pytest.raises(ValueError, match="images do not match"):
            im.paste(im.copy(), box)

    def test_incorrect_abbreviated_form(self) -> None:
        im = Image.new("L", (1, 1))
        with pytest.raises(ValueError):
            im.paste(im, im, im)



"""
Helper functions.
"""

from __future__ import annotations

import logging
import os
import shutil
import subprocess
import sys
import tempfile
from functools import lru_cache
from io import BytesIO

import pytest
from packaging.version import parse as parse_version

from PIL import Image, ImageFile, ImageMath, features

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from pathlib import Path
    from typing import Any

logger = logging.getLogger(__name__)

uploader = None
if os.environ.get("SHOW_ERRORS"):
    uploader = "show"
elif "GITHUB_ACTIONS" in os.environ:
    uploader = "github_actions"


def upload(a: Image.Image, b: Image.Image) -> str | None:
    if uploader == "show":
        # local img.show for errors.
        a.show()
        b.show()
    elif uploader == "github_actions":
        dir_errors = os.path.join(os.path.dirname(__file__), "errors")
        os.makedirs(dir_errors, exist_ok=True)
        tmpdir = tempfile.mkdtemp(dir=dir_errors)
        a.save(os.path.join(tmpdir, "a.png"))
        b.save(os.path.join(tmpdir, "b.png"))
        return tmpdir
    return None


def convert_to_comparable(
    a: Image.Image, b: Image.Image
) -> tuple[Image.Image, Image.Image]:
    new_a, new_b = a, b
    if a.mode == "P":
        new_a = Image.new("L", a.size)
        new_b = Image.new("L", b.size)
        new_a.putdata(a.get_flattened_data())
        new_b.putdata(b.get_flattened_data())
    elif a.mode == "I;16":
        new_a = a.convert("I")
        new_b = b.convert("I")
    return new_a, new_b


def assert_deep_equal(a: Any, b: Any, msg: str | None = None) -> None:
    try:
        assert len(a) == len(b), msg or f"got length {len(a)}, expected {len(b)}"
    except Exception:
        assert a == b, msg


def assert_image(
    im: Image.Image, mode: str, size: tuple[int, int], msg: str | None = None
) -> None:
    if mode is not None:
        assert im.mode == mode, (
            msg or f"got mode {repr(im.mode)}, expected {repr(mode)}"
        )

    if size is not None:
        assert im.size == size, (
            msg or f"got size {repr(im.size)}, expected {repr(size)}"
        )


def assert_image_equal(a: Image.Image, b: Image.Image, msg: str | None = None) -> None:
    assert a.mode == b.mode, msg or f"got mode {repr(a.mode)}, expected {repr(b.mode)}"
    assert a.size == b.size, msg or f"got size {repr(a.size)}, expected {repr(b.size)}"
    if a.tobytes() != b.tobytes():
        try:
            url = upload(a, b)
            if url:
                logger.error("URL for test images: %s", url)
        except Exception:
            pass

        pytest.fail(msg or "got different content")


def assert_image_equal_tofile(
    a: Image.Image,
    filename: str | Path,
    msg: str | None = None,
    mode: str | None = None,
) -> None:
    with Image.open(filename) as im:
        converted_im = im.convert(mode) if mode else im
        assert_image_equal(a, converted_im, msg)


def assert_image_similar(
    a: Image.Image, b: Image.Image, epsilon: float, msg: str | None = None
) -> None:
    assert a.mode == b.mode, msg or f"got mode {repr(a.mode)}, expected {repr(b.mode)}"
    assert a.size == b.size, msg or f"got size {repr(a.size)}, expected {repr(b.size)}"

    a, b = convert_to_comparable(a, b)

    diff = 0
    for ach, bch in zip(a.split(), b.split()):
        chdiff = ImageMath.lambda_eval(
            lambda args: abs(args["a"] - args["b"]), a=ach, b=bch
        ).convert("L")
        diff += sum(i * num for i, num in enumerate(chdiff.histogram()))

    ave_diff = diff / (a.size[0] * a.size[1])
    try:
        assert epsilon >= ave_diff, (
            (msg or "")
            + f" average pixel value difference {ave_diff:.4f} > epsilon {epsilon:.4f}"
        )
    except Exception as e:
        try:
            url = upload(a, b)
            if url:
                logger.exception("URL for test images: %s", url)
        except Exception:
            pass
        raise e


def assert_image_similar_tofile(
    a: Image.Image,
    filename: str | Path,
    epsilon: float,
    msg: str | None = None,
) -> None:
    with Image.open(filename) as img:
        assert_image_similar(a, img, epsilon, msg)


def assert_not_all_same(items: Sequence[Any], msg: str | None = None) -> None:
    assert items.count(items[0]) != len(items), msg


def assert_tuple_approx_equal(
    actuals: Sequence[int], targets: tuple[int, ...], threshold: int, msg: str
) -> None:
    """Tests if actuals has values within threshold from targets"""
    for i, target in enumerate(targets):
        if not (target - threshold <= actuals[i] <= target + threshold):
            pytest.fail(msg + ": " + repr(actuals) + " != " + repr(targets))


def timeout_unless_slower_valgrind(timeout: float) -> pytest.MarkDecorator:
    if "PILLOW_VALGRIND_TEST" in os.environ:
        return pytest.mark.pil_noop_mark()
    return pytest.mark.timeout(timeout)


def skip_unless_feature(feature: str) -> pytest.MarkDecorator:
    reason = f"{feature} not available"
    return pytest.mark.skipif(not features.check(feature), reason=reason)


def has_feature_version(feature: str, required: str) -> bool:
    version = features.version(feature)
    assert version is not None
    version_required = parse_version(required)
    version_available = parse_version(version)
    return version_available >= version_required


def skip_unless_feature_version(
    feature: str, required: str, reason: str | None = None
) -> pytest.MarkDecorator:
    version = features.version(feature)
    if version is None:
        return pytest.mark.skip(f"{feature} not available")
    if reason is None:
        reason = f"{feature} is older than {required}"
    version_required = parse_version(required)
    version_available = parse_version(version)
    return pytest.mark.skipif(version_available < version_required, reason=reason)


def mark_if_feature_version(
    mark: pytest.MarkDecorator,
    feature: str,
    version_blacklist: str,
    reason: str | None = None,
) -> pytest.MarkDecorator:
    version = features.version(feature)
    if version is None:
        return pytest.mark.pil_noop_mark()
    if reason is None:
        reason = f"{feature} is {version_blacklist}"
    version_required = parse_version(version_blacklist)
    version_available = parse_version(version)
    if (
        version_available.major == version_required.major
        and version_available.minor == version_required.minor
    ):
        return mark(reason=reason)
    return pytest.mark.pil_noop_mark()


@pytest.mark.skipif(sys.platform.startswith("win32"), reason="Requires Unix or macOS")
class PillowLeakTestCase:
    # requires unix/macOS
    iterations = 100  # count
    mem_limit = 512  # k

    def _get_mem_usage(self) -> float:
        """
        Gets the RUSAGE memory usage, returns in K. Encapsulates the difference
        between macOS and Linux rss reporting

        :returns: memory usage in kilobytes
        """

        from resource import RUSAGE_SELF, getrusage

        mem = getrusage(RUSAGE_SELF).ru_maxrss
        # man 2 getrusage:
        #     ru_maxrss
        # This is the maximum resident set size utilized
        # in bytes on macOS, in kilobytes on Linux
        return mem / 1024 if sys.platform == "darwin" else mem

    def _test_leak(self, core: Callable[[], None]) -> None:
        start_mem = self._get_mem_usage()
        for cycle in range(self.iterations):
            core()
            mem = self._get_mem_usage() - start_mem
            msg = f"memory usage limit exceeded in iteration {cycle}"
            assert mem < self.mem_limit, msg


# helpers


def fromstring(data: bytes) -> ImageFile.ImageFile:
    return Image.open(BytesIO(data))


def tostring(im: Image.Image, string_format: str, **options: Any) -> bytes:
    out = BytesIO()
    im.save(out, string_format, **options)
    return out.getvalue()


def hopper(mode: str | None = None) -> Image.Image:
    # Use caching to reduce reading from disk, but return a copy
    # so that the cached image isn't modified by the tests
    # (for fast, isolated, repeatable tests).

    if mode is None:
        # Always return fresh not-yet-loaded version of image.
        # Operations on not-yet-loaded images are a separate class of errors
        # that we should catch.
        return Image.open("Tests/images/hopper.ppm")

    return _cached_hopper(mode).copy()


@lru_cache
def _cached_hopper(mode: str) -> Image.Image:
    if mode == "F":
        im = hopper("L")
    else:
        im = hopper()
    try:
        im = im.convert(mode)
    except ImportError:
        if mode == "LAB":
            im = Image.open("Tests/images/hopper.Lab.tif")
        else:
            raise
    return im


def djpeg_available() -> bool:
    if shutil.which("djpeg"):
        try:
            subprocess.check_call(["djpeg", "-version"])
            return True
        except subprocess.CalledProcessError:  # pragma: no cover
            return False
    return False


def netpbm_available() -> bool:
    return bool(shutil.which("ppmquant") and shutil.which("ppmtogif"))


def magick_command() -> list[str] | None:
    if sys.platform == "win32":
        magickhome = os.environ.get("MAGICK_HOME")
        if magickhome:
            imagemagick = [os.path.join(magickhome, "convert.exe")]
            graphicsmagick = [os.path.join(magickhome, "gm.exe"), "convert"]
        else:
            imagemagick = None
            graphicsmagick = None
    else:
        imagemagick = ["convert"]
        graphicsmagick = ["gm", "convert"]

    if imagemagick and shutil.which(imagemagick[0]):
        return imagemagick
    if graphicsmagick and shutil.which(graphicsmagick[0]):
        return graphicsmagick
    return None


def on_ci() -> bool:
    return "CI" in os.environ


def is_big_endian() -> bool:
    return sys.byteorder == "big"


def is_ppc64le() -> bool:
    import platform

    return platform.machine() == "ppc64le"


def is_win32() -> bool:
    return sys.platform.startswith("win32")


def is_pypy() -> bool:
    return hasattr(sys, "pypy_translation_info")


class CachedProperty:
    def __init__(self, func: Callable[[Any], Any]) -> None:
        self.func = func

    def __get__(self, instance: Any, cls: type[Any] | None = None) -> Any:
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result



from __future__ import annotations

import sys

import pytest

from PIL import Image


def test_overflow(monkeypatch: pytest.MonkeyPatch) -> None:
    # There is the potential to overflow comparisons in map.c
    # if there are > SIZE_MAX bytes in the image or if
    # the file encodes an offset that makes
    # (offset + size(bytes)) > SIZE_MAX

    # Note that this image triggers the decompression bomb warning:
    monkeypatch.setattr(Image, "MAX_IMAGE_PIXELS", None)

    # This image hits the offset test.
    with Image.open("Tests/images/l2rgb_read.bmp") as im:
        with pytest.raises((ValueError, MemoryError, OSError)):
            im.load()


def test_tobytes(monkeypatch: pytest.MonkeyPatch) -> None:
    # Note that this image triggers the decompression bomb warning:
    monkeypatch.setattr(Image, "MAX_IMAGE_PIXELS", None)

    # Previously raised an access violation on Windows
    with Image.open("Tests/images/l2rgb_read.bmp") as im:
        with pytest.raises((ValueError, MemoryError, OSError)):
            im.tobytes()


@pytest.mark.skipif(sys.maxsize <= 2**32, reason="Requires 64-bit system")
def test_ysize() -> None:
    numpy = pytest.importorskip("numpy", reason="NumPy not installed")

    # Should not raise 'Integer overflow in ysize'
    arr = numpy.zeros((46341, 46341), dtype=numpy.uint8)
    Image.fromarray(arr)



from __future__ import annotations

import argparse
import json
import os
import platform
import re
import shutil
import struct
import subprocess
import sys
from pathlib import Path
from typing import Any


def cmd_cd(path: str) -> str:
    return f"cd /D {path}"


def cmd_set(name: str, value: str) -> str:
    return f"set {name}={value}"


def cmd_append(name: str, value: str) -> str:
    op = "path " if name == "PATH" else f"set {name}="
    return op + f"%{name}%;{value}"


def cmd_copy(src: str, tgt: str) -> str:
    return f'copy /Y /B "{src}" "{tgt}"'


def cmd_xcopy(src: str, tgt: str) -> str:
    return f'xcopy /Y /E "{src}" "{tgt}"'


def cmd_mkdir(path: str) -> str:
    return f'mkdir "{path}"'


def cmd_rmdir(path: str) -> str:
    return f'rmdir /S /Q "{path}"'


def cmd_nmake(
    makefile: str | None = None,
    target: str = "",
    params: list[str] | None = None,
) -> str:
    return " ".join(
        [
            "{nmake}",
            "-nologo",
            f'-f "{makefile}"' if makefile is not None else "",
            f'{" ".join(params)}' if params is not None else "",
            f'"{target}"',
        ]
    )


def cmds_cmake(
    target: str | tuple[str, ...] | list[str],
    *params: str,
    build_dir: str = ".",
    build_type: str = "Release",
) -> list[str]:
    if not isinstance(target, str):
        target = " ".join(target)

    return [
        " ".join(
            [
                "{cmake}",
                f"-DCMAKE_BUILD_TYPE={build_type}",
                "-DCMAKE_VERBOSE_MAKEFILE=ON",
                "-DCMAKE_RULE_MESSAGES:BOOL=OFF",  # for NMake
                "-DCMAKE_C_COMPILER=cl.exe",  # for Ninja
                "-DCMAKE_CXX_COMPILER=cl.exe",  # for Ninja
                "-DCMAKE_C_FLAGS=-nologo",
                "-DCMAKE_CXX_FLAGS=-nologo",
                *params,
                '-G "{cmake_generator}"',
                f'-B "{build_dir}"',
                "-S .",
            ]
        ),
        f'{{cmake}} --build "{build_dir}" --clean-first --parallel --target {target}',
    ]


def cmd_msbuild(
    file: str,
    configuration: str = "Release",
    target: str = "Build",
    plat: str = "{msbuild_arch}",
) -> str:
    return " ".join(
        [
            "{msbuild}",
            f"{file}",
            f'/t:"{target}"',
            f'/p:Configuration="{configuration}"',
            f"/p:Platform={plat}",
            "/m",
        ]
    )


SF_PROJECTS = "https://sourceforge.net/projects"

ARCHITECTURES = {
    "x86": {"vcvars_arch": "x86", "msbuild_arch": "Win32"},
    "AMD64": {"vcvars_arch": "x86_amd64", "msbuild_arch": "x64"},
    "ARM64": {"vcvars_arch": "x86_arm64", "msbuild_arch": "ARM64"},
}

V = json.loads(
    (Path(__file__).parents[1] / ".github" / "dependencies.json").read_text()
)
V["libpng-xy"] = "".join(V["libpng"].split(".")[:2])


# dependencies, listed in order of compilation
DEPS: dict[str, dict[str, Any]] = {
    "libjpeg": {
        "url": f"https://github.com/libjpeg-turbo/libjpeg-turbo/releases/download/{V['jpegturbo']}/libjpeg-turbo-{V['jpegturbo']}.tar.gz",
        "filename": f"libjpeg-turbo-{V['jpegturbo']}.tar.gz",
        "license": ["README.ijg", "LICENSE.md"],
        "license_pattern": (
            "(LEGAL ISSUES\n============\n\n.+?)\n\nREFERENCES\n=========="
            ".+(libjpeg-turbo Licenses\n======================\n\n.+)$"
        ),
        "patch": {
            r"CMakeLists.txt": {
                # libjpeg-turbo does not detect MSVC x86_arm64 cross-compiler correctly
                'if(MSVC_IDE AND CMAKE_GENERATOR_PLATFORM MATCHES "arm64")': "if({architecture} STREQUAL ARM64)",  # noqa: E501
            },
        },
        "build": [
            *cmds_cmake(
                ("jpeg-static", "djpeg-static"),
                "-DENABLE_SHARED:BOOL=FALSE",
                "-DWITH_JPEG8:BOOL=TRUE",
                "-DWITH_CRT_DLL:BOOL=TRUE",
            ),
            cmd_copy("jpeg-static.lib", "libjpeg.lib"),
            cmd_copy("djpeg-static.exe", "djpeg.exe"),
        ],
        "headers": ["jconfig.h", r"src\j*.h"],
        "libs": ["libjpeg.lib"],
        "bins": ["djpeg.exe"],
    },
    "zlib": {
        "url": f"https://github.com/zlib-ng/zlib-ng/archive/refs/tags/{V['zlib-ng']}.tar.gz",
        "filename": f"zlib-ng-{V['zlib-ng']}.tar.gz",
        "license": "LICENSE.md",
        "patch": {
            r"CMakeLists.txt": {
                "set_target_properties(zlib-ng PROPERTIES OUTPUT_NAME zlibstatic${{SUFFIX}})": "set_target_properties(zlib-ng PROPERTIES OUTPUT_NAME zlib)",  # noqa: E501
            },
        },
        "build": [
            *cmds_cmake(
                "zlib-ng", "-DBUILD_SHARED_LIBS:BOOL=OFF", "-DZLIB_COMPAT:BOOL=ON"
            ),
        ],
        "headers": [r"z*.h"],
        "libs": [r"zlib.lib"],
    },
    "xz": {
        "url": f"https://github.com/tukaani-project/xz/releases/download/v{V['xz']}/FILENAME",
        "filename": f"xz-{V['xz']}.tar.gz",
        "license": "COPYING",
        "build": [
            *cmds_cmake("liblzma", "-DBUILD_SHARED_LIBS:BOOL=OFF"),
            cmd_mkdir(r"{inc_dir}\lzma"),
            cmd_copy(r"src\liblzma\api\lzma\*.h", r"{inc_dir}\lzma"),
        ],
        "headers": [r"src\liblzma\api\lzma.h"],
        "libs": [r"lzma.lib"],
    },
    "libwebp": {
        "url": "http://downloads.webmproject.org/releases/webp/FILENAME",
        "filename": f"libwebp-{V['libwebp']}.tar.gz",
        "license": "COPYING",
        "patch": {
            r"src\enc\picture_csp_enc.c": {
                # link against libsharpyuv.lib
                '#include "sharpyuv/sharpyuv.h"': '#include "sharpyuv/sharpyuv.h"\n#pragma comment(lib, "libsharpyuv.lib")',  # noqa: E501
            }
        },
        "build": [
            *cmds_cmake(
                "webp webpmux webpdemux",
                "-DBUILD_SHARED_LIBS:BOOL=OFF",
                "-DWEBP_LINK_STATIC:BOOL=OFF",
            ),
            cmd_mkdir(r"{inc_dir}\webp"),
            cmd_copy(r"src\webp\*.h", r"{inc_dir}\webp"),
        ],
        "libs": [r"libsharpyuv.lib", r"libwebp*.lib"],
    },
    "libtiff": {
        "url": "https://download.osgeo.org/libtiff/FILENAME",
        "filename": f"tiff-{V['tiff']}.tar.gz",
        "license": "LICENSE.md",
        "patch": {
            r"libtiff\tif_lzma.c": {
                # link against lzma.lib
                "#ifdef LZMA_SUPPORT": '#ifdef LZMA_SUPPORT\n#pragma comment(lib, "lzma.lib")',  # noqa: E501
            },
            r"libtiff\tif_webp.c": {
                # link against libwebp.lib
                "#ifdef WEBP_SUPPORT": '#ifdef WEBP_SUPPORT\n#pragma comment(lib, "libwebp.lib")',  # noqa: E501
            },
            r"libtiff\tif_getimage.c": {
                "int32_t incr = 3 * w + 4 * toskew;": "const tmsize_t incr = 3 * (tmsize_t)w + 4 * (tmsize_t)toskew;",  # noqa: E501
                "int32_t incr = 2 * toskew + w;": "const tmsize_t incr = 2 * (tmsize_t)toskew + w;",  # noqa: E501
            },
        },
        "build": [
            *cmds_cmake(
                "tiff",
                "-DBUILD_SHARED_LIBS:BOOL=OFF",
                "-DWebP_LIBRARY=libwebp",
                '-DCMAKE_C_FLAGS="-nologo -DLZMA_API_STATIC"',
            )
        ],
        "headers": [r"libtiff\tiff*.h"],
        "libs": [r"libtiff\*.lib"],
    },
    "libpng": {
        "url": f"{SF_PROJECTS}/libpng/files/libpng{V['libpng-xy']}/{V['libpng']}/"
        f"FILENAME/download",
        "filename": f"libpng-{V['libpng']}.tar.gz",
        "license": "LICENSE",
        "build": [
            *cmds_cmake("png_static", "-DPNG_SHARED:BOOL=OFF", "-DPNG_TESTS:BOOL=OFF"),
            cmd_copy(
                f"libpng{V['libpng-xy']}_static.lib", f"libpng{V['libpng-xy']}.lib"
            ),
        ],
        "headers": [r"png*.h"],
        "libs": [f"libpng{V['libpng-xy']}.lib"],
    },
    "brotli": {
        "url": f"https://github.com/google/brotli/archive/refs/tags/v{V['brotli']}.tar.gz",
        "filename": f"brotli-{V['brotli']}.tar.gz",
        "license": "LICENSE",
        "build": [
            *cmds_cmake(("brotlicommon", "brotlidec"), "-DBUILD_SHARED_LIBS:BOOL=OFF"),
            cmd_xcopy(r"c\include", "{inc_dir}"),
        ],
        "libs": ["*.lib"],
    },
    "freetype": {
        "url": "https://download.savannah.gnu.org/releases/freetype/FILENAME",
        "filename": f"freetype-{V['freetype']}.tar.gz",
        "license": ["LICENSE.TXT", r"docs\FTL.TXT", r"docs\GPLv2.TXT"],
        "patch": {
            r"builds\windows\vc2010\freetype.vcxproj": {
                # freetype setting is /MD for .dll and /MT for .lib, we need /MD
                "<RuntimeLibrary>MultiThreaded</RuntimeLibrary>": "<RuntimeLibrary>MultiThreadedDLL</RuntimeLibrary>",  # noqa: E501
                # freetype doesn't specify SDK version, MSBuild may guess incorrectly
                '<PropertyGroup Label="Globals">': '<PropertyGroup Label="Globals">\n    <WindowsTargetPlatformVersion>$(WindowsSDKVersion)</WindowsTargetPlatformVersion>',  # noqa: E501
            },
            r"builds\windows\vc2010\freetype.user.props": {
                "<UserDefines></UserDefines>": "<UserDefines>FT_CONFIG_OPTION_SYSTEM_ZLIB;FT_CONFIG_OPTION_USE_PNG;FT_CONFIG_OPTION_USE_HARFBUZZ;FT_CONFIG_OPTION_USE_BROTLI</UserDefines>",  # noqa: E501
                "<UserIncludeDirectories></UserIncludeDirectories>": r"<UserIncludeDirectories>{dir_harfbuzz}\src;{inc_dir}</UserIncludeDirectories>",  # noqa: E501
                "<UserLibraryDirectories></UserLibraryDirectories>": "<UserLibraryDirectories>{lib_dir}</UserLibraryDirectories>",  # noqa: E501
                "<UserDependencies></UserDependencies>": f"<UserDependencies>zlib.lib;libpng{V['libpng-xy']}.lib;brotlicommon.lib;brotlidec.lib</UserDependencies>",  # noqa: E501
            },
            r"src/autofit/afshaper.c": {
                # link against harfbuzz.lib
                "#ifdef FT_CONFIG_OPTION_USE_HARFBUZZ": '#ifdef FT_CONFIG_OPTION_USE_HARFBUZZ\n#pragma comment(lib, "harfbuzz.lib")',  # noqa: E501
            },
        },
        "build": [
            cmd_rmdir("objs"),
            cmd_msbuild(
                r"builds\windows\vc2010\freetype.vcxproj", "Release Static", "Clean"
            ),
            cmd_msbuild(
                r"builds\windows\vc2010\freetype.vcxproj", "Release Static", "Build"
            ),
            cmd_xcopy("include", "{inc_dir}"),
        ],
        "libs": [r"objs\{msbuild_arch}\Release Static\freetype.lib"],
    },
    "lcms2": {
        "url": f"{SF_PROJECTS}/lcms/files/lcms/{V['lcms2']}/FILENAME/download",
        "filename": f"lcms2-{V['lcms2']}.tar.gz",
        "license": "LICENSE",
        "patch": {
            r"Projects\VC2022\lcms2_static\lcms2_static.vcxproj": {
                # default is /MD for x86 and /MT for x64, we need /MD always
                "<RuntimeLibrary>MultiThreaded</RuntimeLibrary>": "<RuntimeLibrary>MultiThreadedDLL</RuntimeLibrary>",  # noqa: E501
                # retarget to default toolset (selected by vcvarsall.bat)
                "<PlatformToolset>v143</PlatformToolset>": "<PlatformToolset>$(DefaultPlatformToolset)</PlatformToolset>",  # noqa: E501
                # retarget to latest (selected by vcvarsall.bat)
                "<WindowsTargetPlatformVersion>10.0</WindowsTargetPlatformVersion>": "<WindowsTargetPlatformVersion>$(WindowsSDKVersion)</WindowsTargetPlatformVersion>",  # noqa: E501
            }
        },
        "build": [
            cmd_rmdir("Lib"),
            cmd_rmdir(r"Projects\VC2022\Release"),
            cmd_msbuild(r"Projects\VC2022\lcms2.sln", "Release", "Clean"),
            cmd_msbuild(
                r"Projects\VC2022\lcms2.sln", "Release", "lcms2_static:Rebuild"
            ),
            cmd_xcopy("include", "{inc_dir}"),
        ],
        "libs": [r"Lib\MS\*.lib"],
    },
    "openjpeg": {
        "url": f"https://github.com/uclouvain/openjpeg/archive/v{V['openjpeg']}.tar.gz",
        "filename": f"openjpeg-{V['openjpeg']}.tar.gz",
        "license": "LICENSE",
        "build": [
            *cmds_cmake(
                "openjp2", "-DBUILD_CODEC:BOOL=OFF", "-DBUILD_SHARED_LIBS:BOOL=OFF"
            ),
            cmd_mkdir(rf"{{inc_dir}}\openjpeg-{V['openjpeg']}"),
            cmd_copy(r"src\lib\openjp2\*.h", rf"{{inc_dir}}\openjpeg-{V['openjpeg']}"),
        ],
        "libs": [r"bin\*.lib"],
    },
    "libimagequant": {
        "url": "https://github.com/ImageOptim/libimagequant/archive/{V['libimagequant']}.tar.gz",
        "filename": f"libimagequant-{V['libimagequant']}.tar.gz",
        "license": "COPYRIGHT",
        "build": [
            cmd_cd("imagequant-sys"),
            "cargo build --release",
        ],
        "headers": ["libimagequant.h"],
        "libs": [r"..\target\release\imagequant_sys.lib"],
    },
    "harfbuzz": {
        "url": f"https://github.com/harfbuzz/harfbuzz/releases/download/{V['harfbuzz']}/FILENAME",
        "filename": f"harfbuzz-{V['harfbuzz']}.tar.xz",
        "license": "COPYING",
        "build": [
            *cmds_cmake(
                "harfbuzz",
                "-DHB_HAVE_FREETYPE:BOOL=TRUE",
                '-DCMAKE_CXX_FLAGS="-nologo -d2FH4-"',
            ),
        ],
        "headers": [r"src\*.h"],
        "libs": [r"*.lib"],
    },
    "fribidi": {
        "url": f"https://github.com/fribidi/fribidi/archive/v{V['fribidi']}.zip",
        "filename": f"fribidi-{V['fribidi']}.zip",
        "license": "COPYING",
        "build": [
            cmd_copy(r"COPYING", rf"{{bin_dir}}\fribidi-{V['fribidi']}-COPYING"),
            cmd_copy(r"{winbuild_dir}\fribidi.cmake", r"CMakeLists.txt"),
            # generated tab.i files cannot be cross-compiled
            " ^&^& ".join(
                [
                    "if {architecture}==ARM64 cmd /c call {vcvarsall} x86",
                    *cmds_cmake("fribidi-gen", "-DARCH=x86", build_dir="build_x86"),
                ]
            ),
            *cmds_cmake("fribidi", "-DARCH={architecture}"),
        ],
        "bins": [r"*.dll"],
    },
    "libavif": {
        "url": f"https://github.com/AOMediaCodec/libavif/archive/v{V['libavif']}.tar.gz",
        "filename": f"libavif-{V['libavif']}.tar.gz",
        "license": "LICENSE",
        "build": [
            "rustup update",
            f"{sys.executable} -m pip install meson",
            *cmds_cmake(
                "avif_static",
                "-DBUILD_SHARED_LIBS=OFF",
                "-DAVIF_LIBSHARPYUV=LOCAL",
                "-DAVIF_LIBYUV=LOCAL",
                "-DAVIF_CODEC_AOM=LOCAL",
                "-DCONFIG_AV1_HIGHBITDEPTH=0",
                "-DAVIF_CODEC_AOM_DECODE=OFF",
                "-DAVIF_CODEC_DAV1D=LOCAL",
                "-DCMAKE_INTERPROCEDURAL_OPTIMIZATION=ON",
                build_type="MinSizeRel",
            ),
            cmd_xcopy("include", "{inc_dir}"),
        ],
        "libs": ["avif.lib"],
    },
}


# based on distutils._msvccompiler from CPython 3.7.4
def find_msvs(architecture: str) -> dict[str, str] | None:
    root = os.environ.get("ProgramFiles(x86)") or os.environ.get("ProgramFiles")
    if not root:
        print("Program Files not found")
        return None

    requires = ["-requires", "Microsoft.VisualStudio.Component.VC.Tools.x86.x64"]
    if architecture == "ARM64":
        requires += ["-requires", "Microsoft.VisualStudio.Component.VC.Tools.ARM64"]

    try:
        vspath = (
            subprocess.check_output(
                [
                    os.path.join(
                        root, "Microsoft Visual Studio", "Installer", "vswhere.exe"
                    ),
                    "-latest",
                    "-prerelease",
                    *requires,
                    "-property",
                    "installationPath",
                    "-products",
                    "*",
                ]
            )
            .decode(encoding="mbcs")
            .strip()
        )
    except (subprocess.CalledProcessError, OSError, UnicodeDecodeError):
        print("vswhere not found")
        return None

    if not os.path.isdir(os.path.join(vspath, "VC", "Auxiliary", "Build")):
        print("Visual Studio seems to be missing C compiler")
        return None

    # vs2017
    msbuild = os.path.join(vspath, "MSBuild", "15.0", "Bin", "MSBuild.exe")
    if not os.path.isfile(msbuild):
        # vs2019
        msbuild = os.path.join(vspath, "MSBuild", "Current", "Bin", "MSBuild.exe")
        if not os.path.isfile(msbuild):
            print("Visual Studio MSBuild not found")
            return None

    vcvarsall = os.path.join(vspath, "VC", "Auxiliary", "Build", "vcvarsall.bat")
    if not os.path.isfile(vcvarsall):
        print("Visual Studio vcvarsall not found")
        return None

    return {
        "vs_dir": vspath,
        "msbuild": f'"{msbuild}"',
        "vcvarsall": f'"{vcvarsall}"',
        "nmake": "nmake.exe",  # nmake selected by vcvarsall
    }


def download_dep(url: str, file: str) -> None:
    import urllib.error
    import urllib.request

    ex = None
    for i in range(3):
        try:
            print(f"Fetching {url} (attempt {i + 1})...")
            content = urllib.request.urlopen(url).read()
            with open(file, "wb") as f:
                f.write(content)
            break
        except urllib.error.URLError as e:
            ex = e
    else:
        raise RuntimeError(ex)


def extract_dep(url: str, filename: str, prefs: dict[str, str]) -> None:
    import tarfile
    import zipfile

    depends_dir = prefs["depends_dir"]
    sources_dir = prefs["src_dir"]

    file = os.path.join(depends_dir, filename)
    if not os.path.exists(file):
        # First try our mirror
        mirror_url = (
            f"https://raw.githubusercontent.com/"
            f"python-pillow/pillow-depends/main/{filename}"
        )
        try:
            download_dep(mirror_url, file)
        except RuntimeError as exc:
            # Otherwise try upstream
            print(exc)
            download_dep(url.replace("FILENAME", filename), file)

    print("Extracting " + filename)
    sources_dir_abs = os.path.abspath(sources_dir)
    if filename.endswith(".zip"):
        with zipfile.ZipFile(file) as zf:
            for member in zf.namelist():
                member_abspath = os.path.abspath(os.path.join(sources_dir, member))
                member_prefix = os.path.commonpath([sources_dir_abs, member_abspath])
                if sources_dir_abs != member_prefix:
                    msg = "Attempted Path Traversal in Zip File"
                    raise RuntimeError(msg)
            zf.extractall(sources_dir)
    elif filename.endswith((".tar.gz", ".tar.xz")):
        with tarfile.open(file, "r:xz" if filename.endswith(".xz") else "r:gz") as tgz:
            for member in tgz.getnames():
                member_abspath = os.path.abspath(os.path.join(sources_dir, member))
                member_prefix = os.path.commonpath([sources_dir_abs, member_abspath])
                if sources_dir_abs != member_prefix:
                    msg = "Attempted Path Traversal in Tar File"
                    raise RuntimeError(msg)
            if sys.version_info >= (3, 12):
                tgz.extractall(sources_dir, filter="data")
            else:
                tgz.extractall(sources_dir)
    else:
        msg = "Unknown archive type: " + filename
        raise RuntimeError(msg)


def write_script(
    name: str, lines: list[str], prefs: dict[str, str], verbose: bool
) -> None:
    name = os.path.join(prefs["build_dir"], name)
    lines = [line.format(**prefs) for line in lines]
    print("Writing " + name)
    with open(name, "w", newline="") as f:
        f.write(os.linesep.join(lines))
    if verbose:
        for line in lines:
            print("    " + line)


def get_footer(dep: dict[str, Any]) -> list[str]:
    return (
        [cmd_copy(out, "{inc_dir}") for out in dep.get("headers", [])]
        + [cmd_copy(out, "{lib_dir}") for out in dep.get("libs", [])]
        + [cmd_copy(out, "{bin_dir}") for out in dep.get("bins", [])]
    )


def build_env(prefs: dict[str, str], verbose: bool) -> None:
    lines = [
        "if defined DISTUTILS_USE_SDK goto end",
        cmd_set("INCLUDE", "{inc_dir}"),
        cmd_set("INCLIB", "{lib_dir}"),
        cmd_set("LIB", "{lib_dir}"),
        cmd_append("PATH", "{bin_dir}"),
        "call {vcvarsall} {vcvars_arch}",
        cmd_set("DISTUTILS_USE_SDK", "1"),  # use same compiler to build Pillow
        cmd_set("py_vcruntime_redist", "true"),  # always use /MD, never /MT
        ":end",
        "@echo on",
    ]
    write_script("build_env.cmd", lines, prefs, verbose)


def build_dep(name: str, prefs: dict[str, str], verbose: bool) -> str:
    dep = DEPS[name]
    directory = dep["dir"]
    file = f"build_dep_{name}.cmd"
    license_dir = prefs["license_dir"]
    sources_dir = prefs["src_dir"]

    extract_dep(dep["url"], dep["filename"], prefs)

    licenses = dep["license"]
    if isinstance(licenses, str):
        licenses = [licenses]
    license_text = ""
    for license_file in licenses:
        with open(os.path.join(sources_dir, directory, license_file)) as f:
            license_text += f.read()
    if "license_pattern" in dep:
        match = re.search(dep["license_pattern"], license_text, re.DOTALL)
        assert match is not None
        license_text = "\n".join(match.groups())
    assert len(license_text) > 50
    with open(os.path.join(license_dir, f"{directory}.txt"), "w") as f:
        print(f"Writing license {directory}.txt")
        f.write(license_text)

    for patch_file, patch_list in dep.get("patch", {}).items():
        patch_file = os.path.join(sources_dir, directory, patch_file.format(**prefs))
        with open(patch_file) as f:
            text = f.read()
        for patch_from, patch_to in patch_list.items():
            patch_from = patch_from.format(**prefs)
            patch_to = patch_to.format(**prefs)
            assert patch_from in text
            text = text.replace(patch_from, patch_to)
        with open(patch_file, "w") as f:
            print(f"Patching {patch_file}")
            f.write(text)

    banner = f"Building {name} ({directory})"
    lines = [
        r'call "{build_dir}\build_env.cmd"',
        "@echo " + ("=" * 70),
        f"@echo ==== {banner:<60} ====",
        "@echo " + ("=" * 70),
        cmd_cd(os.path.join(sources_dir, directory)),
        *dep.get("build", []),
        *get_footer(dep),
    ]

    write_script(file, lines, prefs, verbose)
    return file


def build_dep_all(disabled: list[str], prefs: dict[str, str], verbose: bool) -> None:
    lines = [r'call "{build_dir}\build_env.cmd"']
    gha_groups = "GITHUB_ACTIONS" in os.environ
    for dep_name in DEPS:
        print()
        if dep_name in disabled:
            print(f"Skipping disabled dependency {dep_name}")
            continue
        script = build_dep(dep_name, prefs, verbose)
        if gha_groups:
            lines.append(f"@echo ::group::Running {script}")
        lines.append(rf'cmd.exe /c "{{build_dir}}\{script}"')
        lines.append("if errorlevel 1 echo Build failed! && exit /B 1")
        if gha_groups:
            lines.append("@echo ::endgroup::")
    print()
    lines.append("@echo All Pillow dependencies built successfully!")
    write_script("build_dep_all.cmd", lines, prefs, verbose)


def main() -> None:
    winbuild_dir = os.path.dirname(os.path.realpath(__file__))

    parser = argparse.ArgumentParser(
        prog="winbuild\\build_prepare.py",
        description="Download and generate build scripts for Pillow dependencies.",
        epilog="""Arguments can also be supplied using the environment variables
                  PILLOW_BUILD, PILLOW_DEPS, ARCHITECTURE. See winbuild\\build.rst
                  for more information.""",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="print generated scripts"
    )
    parser.add_argument(
        "-d",
        "--dir",
        "--build-dir",
        dest="build_dir",
        metavar="PILLOW_BUILD",
        default=os.environ.get("PILLOW_BUILD", os.path.join(winbuild_dir, "build")),
        help="build directory (default: 'winbuild\\build')",
    )
    parser.add_argument(
        "--depends",
        dest="depends_dir",
        metavar="PILLOW_DEPS",
        default=os.environ.get("PILLOW_DEPS", os.path.join(winbuild_dir, "depends")),
        help="directory used to store cached dependencies "
        "(default: 'winbuild\\depends')",
    )
    parser.add_argument(
        "--architecture",
        choices=ARCHITECTURES,
        default=os.environ.get(
            "ARCHITECTURE",
            (
                "ARM64"
                if platform.machine() == "ARM64"
                else ("x86" if struct.calcsize("P") == 4 else "AMD64")
            ),
        ),
        help="build architecture (default: same as host Python)",
    )
    parser.add_argument(
        "--nmake",
        dest="cmake_generator",
        action="store_const",
        const="NMake Makefiles",
        default="Ninja",
        help="build dependencies using NMake instead of Ninja",
    )
    parser.add_argument(
        "--no-imagequant",
        action="store_true",
        help="skip GPL-licensed optional dependency libimagequant",
    )
    parser.add_argument(
        "--no-fribidi",
        "--no-raqm",
        action="store_true",
        help="skip LGPL-licensed optional dependency FriBiDi",
    )
    parser.add_argument(
        "--no-avif",
        action="store_true",
        help="skip optional dependency libavif",
    )
    args = parser.parse_args()

    arch_prefs = ARCHITECTURES[args.architecture]
    print("Target architecture:", args.architecture)

    msvs = find_msvs(args.architecture)
    if msvs is None:
        msg = "Visual Studio not found. Please install Visual Studio 2017 or newer."
        raise RuntimeError(msg)
    print("Found Visual Studio at:", msvs["vs_dir"])

    # dependency cache directory
    args.depends_dir = os.path.abspath(args.depends_dir)
    os.makedirs(args.depends_dir, exist_ok=True)
    print("Caching dependencies in:", args.depends_dir)

    args.build_dir = os.path.abspath(args.build_dir)
    print("Using output directory:", args.build_dir)

    # build directory for *.h files
    inc_dir = os.path.join(args.build_dir, "inc")
    # build directory for *.lib files
    lib_dir = os.path.join(args.build_dir, "lib")
    # build directory for *.bin files
    bin_dir = os.path.join(args.build_dir, "bin")
    # directory for storing project files
    sources_dir = os.path.join(args.build_dir, "src")
    # copy dependency licenses to this directory
    license_dir = os.path.join(args.build_dir, "license")

    shutil.rmtree(args.build_dir, ignore_errors=True)
    os.makedirs(args.build_dir, exist_ok=False)
    for path in [inc_dir, lib_dir, bin_dir, sources_dir, license_dir]:
        os.makedirs(path, exist_ok=True)

    disabled = []
    if args.no_imagequant:
        disabled += ["libimagequant"]
    if args.no_fribidi:
        disabled += ["fribidi"]
    if args.no_avif or args.architecture == "ARM64":
        disabled += ["libavif"]

    prefs = {
        "architecture": args.architecture,
        **arch_prefs,
        # Pillow paths
        "winbuild_dir": winbuild_dir,
        # Build paths
        "bin_dir": bin_dir,
        "build_dir": args.build_dir,
        "depends_dir": args.depends_dir,
        "inc_dir": inc_dir,
        "lib_dir": lib_dir,
        "license_dir": license_dir,
        "src_dir": sources_dir,
        # Compilers / Tools
        **msvs,
        "cmake": "cmake.exe",  # TODO find CMAKE automatically
        "cmake_generator": args.cmake_generator,
        # TODO find NASM automatically
    }

    for k, v in DEPS.items():
        if "dir" not in v:
            v["dir"] = re.sub(r"\.(tar\.gz|tar\.xz|zip)", "", v["filename"])
        prefs[f"dir_{k}"] = os.path.join(sources_dir, v["dir"])

    print()

    write_script(".gitignore", ["*"], prefs, args.verbose)
    build_env(prefs, args.verbose)
    build_dep_all(disabled, prefs, args.verbose)


if __name__ == "__main__":
    main()



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import GbrImagePlugin, Image, _binary

from .helper import assert_image_equal_tofile


def test_gbr_file() -> None:
    with Image.open("Tests/images/gbr.gbr") as im:
        assert_image_equal_tofile(im, "Tests/images/gbr.png")


def test_load() -> None:
    with Image.open("Tests/images/gbr.gbr") as im:
        px = im.load()
        assert px is not None
        assert px[0, 0] == (0, 0, 0, 0)

        # Test again now that it has already been loaded once
        px = im.load()
        assert px is not None
        assert px[0, 0] == (0, 0, 0, 0)


def test_multiple_load_operations() -> None:
    with Image.open("Tests/images/gbr.gbr") as im:
        im.load()
        im.load()
        assert_image_equal_tofile(im, "Tests/images/gbr.png")


def create_gbr_image(info: dict[str, int] = {}, magic_number: bytes = b"") -> BytesIO:
    return BytesIO(
        b"".join(
            _binary.o32be(i)
            for i in [
                info.get("header_size", 20),
                info.get("version", 1),
                info.get("width", 1),
                info.get("height", 1),
                info.get("color_depth", 1),
            ]
        )
        + magic_number
    )


def test_invalid_file() -> None:
    for f in [
        create_gbr_image({"header_size": 0}),
        create_gbr_image({"width": 0}),
        create_gbr_image({"height": 0}),
    ]:
        with pytest.raises(SyntaxError, match="not a GIMP brush"):
            GbrImagePlugin.GbrImageFile(f)

    invalid_file = "Tests/images/flower.jpg"
    with pytest.raises(SyntaxError, match="Unsupported GIMP brush version"):
        GbrImagePlugin.GbrImageFile(invalid_file)


def test_unsupported_gimp_brush() -> None:
    f = create_gbr_image({"color_depth": 2})
    with pytest.raises(SyntaxError, match="Unsupported GIMP brush color depth: 2"):
        GbrImagePlugin.GbrImageFile(f)


def test_bad_magic_number() -> None:
    f = create_gbr_image({"version": 2}, magic_number=b"badm")
    with pytest.raises(SyntaxError, match="not a GIMP brush, bad magic number"):
        GbrImagePlugin.GbrImageFile(f)


def test_L() -> None:
    f = create_gbr_image()
    with Image.open(f) as im:
        assert im.mode == "L"



from __future__ import annotations

from typing import Any

import pytest
from packaging.version import parse as parse_version

from PIL import Image

from .helper import hopper

numpy = pytest.importorskip("numpy", reason="NumPy not installed")

im = hopper().resize((128, 100))

TYPE_CHECKING = False
if TYPE_CHECKING:
    import numpy.typing as npt


def test_toarray() -> None:
    def test(mode: str) -> tuple[tuple[int, ...], str, int]:
        ai = numpy.array(im.convert(mode))
        return ai.shape, ai.dtype.str, ai.nbytes

    def test_with_dtype(dtype: npt.DTypeLike) -> None:
        ai = numpy.array(im, dtype=dtype)
        assert ai.dtype.type is dtype

    # assert test("1") == ((100, 128), '|b1', 1600))
    assert test("L") == ((100, 128), "|u1", 12800)

    # FIXME: wrong?
    assert test("I") == ((100, 128), Image._ENDIAN + "i4", 51200)
    # FIXME: wrong?
    assert test("F") == ((100, 128), Image._ENDIAN + "f4", 51200)

    assert test("LA") == ((100, 128, 2), "|u1", 25600)
    assert test("RGB") == ((100, 128, 3), "|u1", 38400)
    assert test("RGBA") == ((100, 128, 4), "|u1", 51200)
    assert test("RGBX") == ((100, 128, 4), "|u1", 51200)

    test_with_dtype(numpy.float64)
    test_with_dtype(numpy.uint8)

    with Image.open("Tests/images/truncated_jpeg.jpg") as im_truncated:
        if parse_version(numpy.__version__) >= parse_version("1.23"):
            with pytest.raises(OSError):
                numpy.array(im_truncated)
        else:
            with pytest.warns(DeprecationWarning, match="__array_interface__"):
                numpy.array(im_truncated)


def test_fromarray() -> None:
    class Wrapper:
        """Class with API matching Image.fromarray"""

        def __init__(self, img: Image.Image, arr_params: dict[str, Any]) -> None:
            self.img = img
            self.__array_interface__ = arr_params

        def __len__(self) -> int:
            return len(self.img.tobytes())

        def tobytes(self) -> bytes:
            return self.img.tobytes()

    def test(mode: str) -> tuple[str, tuple[int, int], bool]:
        i = im.convert(mode)
        a = numpy.array(i)
        # Make wrapper instance for image, new array interface
        wrapped = Wrapper(
            i,
            {
                "shape": a.shape,
                "typestr": a.dtype.str,
                "version": 3,
                "data": a.data,
                "strides": 1,  # pretend it's non-contiguous
            },
        )
        out = Image.fromarray(wrapped)
        return out.mode, out.size, i.get_flattened_data() == out.get_flattened_data()

    # assert test("1") == ("1", (128, 100), True)
    assert test("L") == ("L", (128, 100), True)
    assert test("I") == ("I", (128, 100), True)
    assert test("F") == ("F", (128, 100), True)
    assert test("LA") == ("LA", (128, 100), True)
    assert test("RGB") == ("RGB", (128, 100), True)
    assert test("RGBA") == ("RGBA", (128, 100), True)
    assert test("RGBX") == ("RGBA", (128, 100), True)

    # Test mode is None with no "typestr" in the array interface
    wrapped = Wrapper(hopper("L"), {"shape": (100, 128)})
    with pytest.raises(TypeError):
        Image.fromarray(wrapped)


def test_fromarray_strides_without_tobytes() -> None:
    class Wrapper:
        def __init__(self, arr_params: dict[str, Any]) -> None:
            self.__array_interface__ = arr_params

        def __len__(self) -> int:
            return 1

    with pytest.raises(ValueError):
        wrapped = Wrapper({"shape": (1, 1), "strides": (1, 1), "typestr": "|u1"})
        Image.fromarray(wrapped, "L")


def test_fromarray_palette() -> None:
    # Arrange
    i = im.convert("L")
    a = numpy.array(i)

    # Act
    out = Image.fromarray(a, "P")

    # Assert that the Python and C palettes match
    assert out.palette is not None
    assert len(out.palette.colors) == len(out.im.getpalette()) / 3


def test_deprecation() -> None:
    a = numpy.array(im.convert("L"))
    with pytest.warns(
        DeprecationWarning, match="'mode' parameter for changing data types"
    ):
        Image.fromarray(a, "1")



from __future__ import annotations

from PIL import Image

repro_ss2 = (
    "images/fli_oob/06r/06r00.fli",
    "images/fli_oob/06r/others/06r01.fli",
    "images/fli_oob/06r/others/06r02.fli",
    "images/fli_oob/06r/others/06r03.fli",
    "images/fli_oob/06r/others/06r04.fli",
)

repro_lc = (
    "images/fli_oob/05r/05r00.fli",
    "images/fli_oob/05r/others/05r03.fli",
    "images/fli_oob/05r/others/05r06.fli",
    "images/fli_oob/05r/others/05r05.fli",
    "images/fli_oob/05r/others/05r01.fli",
    "images/fli_oob/05r/others/05r04.fli",
    "images/fli_oob/05r/others/05r02.fli",
    "images/fli_oob/05r/others/05r07.fli",
    "images/fli_oob/patch0/000000",
    "images/fli_oob/patch0/000001",
    "images/fli_oob/patch0/000002",
    "images/fli_oob/patch0/000003",
)


repro_advance = (
    "images/fli_oob/03r/03r00.fli",
    "images/fli_oob/03r/others/03r01.fli",
    "images/fli_oob/03r/others/03r09.fli",
    "images/fli_oob/03r/others/03r11.fli",
    "images/fli_oob/03r/others/03r05.fli",
    "images/fli_oob/03r/others/03r10.fli",
    "images/fli_oob/03r/others/03r06.fli",
    "images/fli_oob/03r/others/03r08.fli",
    "images/fli_oob/03r/others/03r03.fli",
    "images/fli_oob/03r/others/03r07.fli",
    "images/fli_oob/03r/others/03r02.fli",
    "images/fli_oob/03r/others/03r04.fli",
)

repro_brun = (
    "images/fli_oob/04r/initial.fli",
    "images/fli_oob/04r/others/04r02.fli",
    "images/fli_oob/04r/others/04r05.fli",
    "images/fli_oob/04r/others/04r04.fli",
    "images/fli_oob/04r/others/04r03.fli",
    "images/fli_oob/04r/others/04r01.fli",
    "images/fli_oob/04r/04r00.fli",
)

repro_copy = (
    "images/fli_oob/02r/others/02r02.fli",
    "images/fli_oob/02r/others/02r04.fli",
    "images/fli_oob/02r/others/02r03.fli",
    "images/fli_oob/02r/others/02r01.fli",
    "images/fli_oob/02r/02r00.fli",
)


for path in repro_ss2 + repro_lc + repro_advance + repro_brun + repro_copy:
    with Image.open(path) as im:
        try:
            im.load()
        except Exception as msg:
            print(msg)



from __future__ import annotations

import io
import os

import pytest

from PIL import Image, SunImagePlugin, _binary

from .helper import assert_image_equal_tofile, assert_image_similar, hopper

EXTRA_DIR = "Tests/images/sunraster"


def test_sanity() -> None:
    # Arrange
    # Created with ImageMagick: convert hopper.jpg hopper.ras
    test_file = "Tests/images/hopper.ras"

    # Act
    with Image.open(test_file) as im:
        # Assert
        assert im.size == (128, 128)

        assert_image_similar(im, hopper(), 5)  # visually verified

    invalid_file = "Tests/images/flower.jpg"
    with pytest.raises(SyntaxError):
        SunImagePlugin.SunImageFile(invalid_file)


def test_im1() -> None:
    with Image.open("Tests/images/sunraster.im1") as im:
        assert_image_equal_tofile(im, "Tests/images/sunraster.im1.png")


def _sun_header(
    depth: int = 0, file_type: int = 0, palette_length: int = 0
) -> io.BytesIO:
    return io.BytesIO(
        _binary.o32be(0x59A66A95)
        + b"\x00" * 8
        + _binary.o32be(depth)
        + b"\x00" * 4
        + _binary.o32be(file_type)
        + b"\x00" * 4
        + _binary.o32be(palette_length)
    )


def test_unsupported_mode_bit_depth() -> None:
    with pytest.raises(SyntaxError, match="Unsupported Mode/Bit Depth"):
        with SunImagePlugin.SunImageFile(_sun_header()):
            pass


def test_unsupported_color_palette_length() -> None:
    with pytest.raises(SyntaxError, match="Unsupported Color Palette Length"):
        with SunImagePlugin.SunImageFile(_sun_header(depth=1, palette_length=1025)):
            pass


def test_unsupported_palette_type() -> None:
    with pytest.raises(SyntaxError, match="Unsupported Palette Type"):
        with SunImagePlugin.SunImageFile(_sun_header(depth=1, palette_length=1)):
            pass


def test_unsupported_file_type() -> None:
    with pytest.raises(SyntaxError, match="Unsupported Sun Raster file type"):
        with SunImagePlugin.SunImageFile(_sun_header(depth=1, file_type=6)):
            pass


@pytest.mark.skipif(
    not os.path.exists(EXTRA_DIR), reason="Extra image files not installed"
)
def test_rgbx() -> None:
    with open(os.path.join(EXTRA_DIR, "32bpp.ras"), "rb") as fp:
        data = fp.read()

    # Set file type to 3
    data = data[:20] + _binary.o32be(3) + data[24:]

    with Image.open(io.BytesIO(data)) as im:
        r, g, b = im.split()
        im_rgb = Image.merge("RGB", (b, g, r))
        assert_image_equal_tofile(im_rgb, os.path.join(EXTRA_DIR, "32bpp.png"))


@pytest.mark.skipif(
    not os.path.exists(EXTRA_DIR), reason="Extra image files not installed"
)
def test_others() -> None:
    files = (
        os.path.join(EXTRA_DIR, f)
        for f in os.listdir(EXTRA_DIR)
        if os.path.splitext(f)[1] in (".sun", ".SUN", ".ras")
    )
    for path in files:
        with Image.open(path) as im:
            im.load()
            assert isinstance(im, SunImagePlugin.SunImageFile)
            assert_image_equal_tofile(im, f"{os.path.splitext(path)[0]}.png")



from __future__ import annotations

import base64
import os

if __name__ == "__main__":
    # create font data chunk for embedding
    font = "Tests/images/courB08"
    print("    f._load_pilfont_data(")
    print(f"         # {os.path.basename(font)}")
    print("         BytesIO(base64.decodestring(b'''")
    with open(font + ".pil", "rb") as fp:
        print(base64.b64encode(fp.read()).decode())
    print("''')), Image.open(BytesIO(base64.decodestring(b'''")
    with open(font + ".pbm", "rb") as fp:
        print(base64.b64encode(fp.read()).decode())
    print("'''))))")



from __future__ import annotations

import warnings
from io import BytesIO
from typing import Any

import pytest

from PIL import (
    Image,
    ImageFile,
    JpegImagePlugin,
    MpoImagePlugin,
    TiffImagePlugin,
    _binary,
)

from .helper import (
    assert_image_equal,
    assert_image_similar,
    is_pypy,
    skip_unless_feature,
)

test_files = ["Tests/images/sugarshack.mpo", "Tests/images/frozenpond.mpo"]

pytestmark = skip_unless_feature("jpg")


def roundtrip(im: Image.Image, **options: Any) -> ImageFile.ImageFile:
    out = BytesIO()
    im.save(out, "MPO", **options)
    out.seek(0)
    return Image.open(out)


@pytest.mark.parametrize("test_file", test_files)
def test_sanity(test_file: str) -> None:
    def check(im: ImageFile.ImageFile) -> None:
        im.load()
        assert im.mode == "RGB"
        assert im.size == (640, 480)
        assert im.format == "MPO"

    with Image.open(test_file) as im:
        check(im)
    with MpoImagePlugin.MpoImageFile(test_file) as im:
        check(im)


@pytest.mark.skipif(is_pypy(), reason="Requires CPython")
def test_unclosed_file() -> None:
    def open_test_image() -> None:
        im = Image.open(test_files[0])
        im.load()

    with pytest.warns(ResourceWarning):
        open_test_image()


def test_closed_file() -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        im = Image.open(test_files[0])
        im.load()
        im.close()


def test_seek_after_close() -> None:
    im = Image.open(test_files[0])
    im.close()

    with pytest.raises(ValueError):
        im.seek(1)


def test_context_manager() -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        with Image.open(test_files[0]) as im:
            im.load()


@pytest.mark.parametrize("test_file", test_files)
def test_app(test_file: str) -> None:
    # Test APP/COM reader (@PIL135)
    with Image.open(test_file) as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        assert im.applist[0][0] == "APP1"
        assert im.applist[1][0] == "APP2"
        assert im.applist[1][1].startswith(
            b"MPF\x00MM\x00*\x00\x00\x00\x08\x00\x03\xb0\x00"
        )
        assert len(im.applist) == 2


@pytest.mark.parametrize("test_file", test_files)
def test_exif(test_file: str) -> None:
    with Image.open(test_file) as im_original:
        im_reloaded = roundtrip(im_original, save_all=True, exif=im_original.getexif())

    for im in (im_original, im_reloaded):
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        info = im._getexif()
        assert info is not None
        assert info[272] == "Nintendo 3DS"
        assert info[296] == 2
        assert info[34665] == 188


def test_frame_size() -> None:
    with Image.open("Tests/images/frame_size.mpo") as im:
        assert im.size == (56, 70)
        im.load()

        im.seek(1)
        assert im.size == (349, 434)
        im.load()

        im.seek(0)
        assert im.size == (56, 70)


def test_ignore_frame_size() -> None:
    # Ignore the different size of the second frame
    # since this is not a "Large Thumbnail" image
    with Image.open("Tests/images/ignore_frame_size.mpo") as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        assert im.size == (64, 64)

        im.seek(1)
        assert im.mpinfo is not None
        assert (
            im.mpinfo[0xB002][1]["Attribute"]["MPType"]
            == "Multi-Frame Image: (Disparity)"
        )
        assert im.size == (64, 64)


def test_parallax() -> None:
    # Nintendo
    with Image.open("Tests/images/sugarshack.mpo") as im:
        exif = im.getexif()
        assert exif.get_ifd(0x927C)[0x1101]["Parallax"] == -44.798187255859375

    # Fujifilm
    with Image.open("Tests/images/fujifilm.mpo") as im:
        im.seek(1)
        exif = im.getexif()
        assert exif.get_ifd(0x927C)[0xB211] == -3.125


def test_truncated_makernote() -> None:
    def check(ifd: TiffImagePlugin.ImageFileDirectory_v2) -> None:
        fp = BytesIO()
        ifd.save(fp)

        e = Image.Exif()
        e.load(fp.getvalue())
        assert e.get_ifd(37500) == {}

    # Nintendo
    ifd = TiffImagePlugin.ImageFileDirectory_v2()
    ifd[271] = "Nintendo"
    ifd[34665] = {37500: b" "}
    check(ifd)

    # Fujifilm
    for data in (
        b"FUJIFILM",
        b"FUJIFILM" + _binary.o32le(50),
        b"FUJIFILM" + _binary.o32le(0),
    ):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        ifd[34665] = {37500: data}
        check(ifd)


def test_reload_exif_after_seek() -> None:
    with Image.open("Tests/images/sugarshack.mpo") as im:
        exif = im.getexif()
        del exif[296]

        im.seek(1)
        assert 296 in exif


@pytest.mark.parametrize("test_file", test_files)
def test_mp(test_file: str) -> None:
    with Image.open(test_file) as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        mpinfo = im._getmp()
        assert mpinfo is not None
        assert mpinfo[45056] == b"0100"
        assert mpinfo[45057] == 2


def test_mp_offset() -> None:
    # This image has been manually hexedited to have an IFD offset of 10
    # in APP2 data, in contrast to normal 8
    with Image.open("Tests/images/sugarshack_ifd_offset.mpo") as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        mpinfo = im._getmp()
        assert mpinfo is not None
        assert mpinfo[45056] == b"0100"
        assert mpinfo[45057] == 2


def test_mp_no_data() -> None:
    # This image has been manually hexedited to have the second frame
    # beyond the end of the file
    with Image.open("Tests/images/sugarshack_no_data.mpo") as im:
        with pytest.raises(ValueError):
            im.seek(1)


@pytest.mark.parametrize("test_file", test_files)
def test_mp_attribute(test_file: str) -> None:
    with Image.open(test_file) as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        mpinfo = im._getmp()
    assert mpinfo is not None
    for frame_number, mpentry in enumerate(mpinfo[0xB002]):
        mpattr = mpentry["Attribute"]
        if frame_number:
            assert not mpattr["RepresentativeImageFlag"]
        else:
            assert mpattr["RepresentativeImageFlag"]
        assert not mpattr["DependentParentImageFlag"]
        assert not mpattr["DependentChildImageFlag"]
        assert mpattr["ImageDataFormat"] == "JPEG"
        assert mpattr["MPType"] == "Multi-Frame Image: (Disparity)"
        assert mpattr["Reserved"] == 0


@pytest.mark.parametrize("test_file", test_files)
def test_seek(test_file: str) -> None:
    with Image.open(test_file) as im:
        assert im.tell() == 0
        # prior to first image raises an error, both blatant and borderline
        with pytest.raises(EOFError):
            im.seek(-1)
        with pytest.raises(EOFError):
            im.seek(-523)
        # after the final image raises an error,
        # both blatant and borderline
        with pytest.raises(EOFError):
            im.seek(2)
        with pytest.raises(EOFError):
            im.seek(523)
        # bad calls shouldn't change the frame
        assert im.tell() == 0
        # this one will work
        im.seek(1)
        assert im.tell() == 1
        # and this one, too
        im.seek(0)
        assert im.tell() == 0


def test_n_frames() -> None:
    with Image.open("Tests/images/sugarshack.mpo") as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        assert im.n_frames == 2
        assert im.is_animated


def test_eoferror() -> None:
    with Image.open("Tests/images/sugarshack.mpo") as im:
        assert isinstance(im, MpoImagePlugin.MpoImageFile)
        n_frames = im.n_frames

        # Test seeking past the last frame
        with pytest.raises(EOFError):
            im.seek(n_frames)
        assert im.tell() < n_frames

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)


def test_adopt_jpeg() -> None:
    with Image.open("Tests/images/hopper.jpg") as im:
        assert isinstance(im, JpegImagePlugin.JpegImageFile)

        with pytest.raises(ValueError):
            MpoImagePlugin.MpoImageFile.adopt(im)


def test_ultra_hdr() -> None:
    with Image.open("Tests/images/ultrahdr.jpg") as im:
        assert im.format == "JPEG"


@pytest.mark.parametrize("test_file", test_files)
def test_image_grab(test_file: str) -> None:
    with Image.open(test_file) as im:
        assert im.tell() == 0
        im0 = im.tobytes()
        im.seek(1)
        assert im.tell() == 1
        im1 = im.tobytes()
        im.seek(0)
        assert im.tell() == 0
        im02 = im.tobytes()
        assert im0 == im02
        assert im0 != im1


@pytest.mark.parametrize("test_file", test_files)
def test_save(test_file: str) -> None:
    with Image.open(test_file) as im:
        assert im.tell() == 0
        jpg0 = roundtrip(im)
        assert_image_similar(im, jpg0, 30)
        im.seek(1)
        assert im.tell() == 1
        jpg1 = roundtrip(im)
        assert_image_similar(im, jpg1, 30)


def test_save_all() -> None:
    for test_file in test_files:
        with Image.open(test_file) as im:
            im_reloaded = roundtrip(im, save_all=True)

            im.seek(0)
            assert_image_similar(im, im_reloaded, 30)

            im.seek(1)
            im_reloaded.seek(1)
            assert_image_similar(im, im_reloaded, 30)

    im_rgb = Image.new("RGB", (1, 1))
    for colors in (("#f00",), ("#f00", "#0f0")):
        append_images = [Image.new("RGB", (1, 1), color) for color in colors]
        im_reloaded = roundtrip(im_rgb, save_all=True, append_images=append_images)

        assert_image_equal(im_rgb, im_reloaded)
        assert isinstance(im_reloaded, MpoImagePlugin.MpoImageFile)
        assert im_reloaded.mpinfo is not None
        assert im_reloaded.mpinfo[45056] == b"0100"

        for im_expected in append_images:
            im_reloaded.seek(im_reloaded.tell() + 1)
            assert_image_similar(im_reloaded, im_expected, 1)

    # Test that a single frame image will not be saved as an MPO
    jpg = roundtrip(im_rgb, save_all=True)
    assert "mp" not in jpg.info


def test_save_xmp() -> None:
    im = Image.new("RGB", (1, 1))
    im2 = Image.new("RGB", (1, 1), "#f00")

    def roundtrip_xmp() -> list[Any]:
        im_reloaded = roundtrip(im, xmp=b"Default", save_all=True, append_images=[im2])
        xmp = [im_reloaded.info["xmp"]]
        im_reloaded.seek(1)
        return xmp + [im_reloaded.info["xmp"]]

    # Use the save parameters for all frames by default
    assert roundtrip_xmp() == [b"Default", b"Default"]

    # Specify a value for the first frame
    im.encoderinfo = {"xmp": b"First frame"}
    assert roundtrip_xmp() == [b"First frame", b"Default"]
    del im.encoderinfo

    # Specify value for the second frame
    im2.encoderinfo = {"xmp": b"Second frame"}
    assert roundtrip_xmp() == [b"Default", b"Second frame"]

    # Test that encoderinfo is unchanged
    assert im2.encoderinfo == {"xmp": b"Second frame"}



from __future__ import annotations

from io import BytesIO
from typing import Any

import pytest

from PIL import (
    BmpImagePlugin,
    EpsImagePlugin,
    Image,
    ImageFile,
    UnidentifiedImageError,
    _binary,
    features,
)

from .helper import (
    assert_image,
    assert_image_equal,
    assert_image_similar,
    fromstring,
    hopper,
    skip_unless_feature,
    tostring,
)

# save original block sizes
MAXBLOCK = ImageFile.MAXBLOCK
SAFEBLOCK = ImageFile.SAFEBLOCK


class TestImageFile:
    def test_parser(self, monkeypatch: pytest.MonkeyPatch) -> None:
        def roundtrip(format: str) -> tuple[Image.Image, Image.Image]:
            im = hopper("L").resize((1000, 1000), Image.Resampling.NEAREST)
            if format in ("MSP", "XBM"):
                im = im.convert("1")

            test_file = BytesIO()

            im.copy().save(test_file, format)

            data = test_file.getvalue()

            parser = ImageFile.Parser()
            parser.feed(data)
            im_out = parser.close()

            return im, im_out

        assert_image_equal(*roundtrip("BMP"))
        im1, im2 = roundtrip("GIF")
        assert_image_similar(im1.convert("P"), im2, 1)
        assert_image_equal(*roundtrip("IM"))
        assert_image_equal(*roundtrip("MSP"))
        if features.check("zlib"):
            # force multiple blocks in PNG driver
            monkeypatch.setattr(ImageFile, "MAXBLOCK", 8192)
            assert_image_equal(*roundtrip("PNG"))
        assert_image_equal(*roundtrip("PPM"))
        assert_image_equal(*roundtrip("TIFF"))
        assert_image_equal(*roundtrip("XBM"))
        assert_image_equal(*roundtrip("TGA"))
        assert_image_equal(*roundtrip("PCX"))

        if EpsImagePlugin.has_ghostscript():
            im1, im2 = roundtrip("EPS")
            # This test fails on Ubuntu 12.04, PPC (Bigendian) It
            # appears to be a ghostscript 9.05 bug, since the
            # ghostscript rendering is wonky and the file is identical
            # to that written on ubuntu 12.04 x64
            # md5sum: ba974835ff2d6f3f2fd0053a23521d4a

            # EPS comes back in RGB:
            assert_image_similar(im1, im2.convert("L"), 20)

        if features.check("jpg"):
            im1, im2 = roundtrip("JPEG")  # lossy compression
            assert_image(im1, im2.mode, im2.size)

        with pytest.raises(OSError):
            roundtrip("PDF")

    def test_ico(self) -> None:
        with open("Tests/images/python.ico", "rb") as f:
            data = f.read()
        with ImageFile.Parser() as p:
            p.feed(data)
            assert p.image is not None
            assert (48, 48) == p.image.size

    @pytest.mark.filterwarnings("ignore:Corrupt EXIF data")
    def test_incremental_tiff(self) -> None:
        with ImageFile.Parser() as p:
            with open("Tests/images/hopper.tif", "rb") as f:
                p.feed(f.read(1024))

                # Check that insufficient data was given in the first feed
                assert not p.image

                p.feed(f.read())
            assert p.image is not None
            assert (128, 128) == p.image.size

    @skip_unless_feature("webp")
    def test_incremental_webp(self) -> None:
        with ImageFile.Parser() as p:
            with open("Tests/images/hopper.webp", "rb") as f:
                p.feed(f.read(1024))

                # Check that insufficient data was given in the first feed
                assert not p.image

                p.feed(f.read())
            assert p.image is not None
            assert (128, 128) == p.image.size

    @skip_unless_feature("zlib")
    def test_safeblock(self, monkeypatch: pytest.MonkeyPatch) -> None:
        im1 = hopper()

        monkeypatch.setattr(ImageFile, "SAFEBLOCK", 1)
        im2 = fromstring(tostring(im1, "PNG"))

        assert_image_equal(im1, im2)

    def test_tile_size(self) -> None:
        with open("Tests/images/hopper.tif", "rb") as im_fp:
            data = im_fp.read()

        reads = []

        class FP(BytesIO):
            def read(self, size: int | None = None) -> bytes:
                reads.append(size)
                return super().read(size)

        fp = FP(data)
        with Image.open(fp) as im:
            assert len(im.tile) == 7

            im.load()

        # Despite multiple tiles, assert only one tile caused a read of maxblock size
        assert reads.count(im.decodermaxblock) == 1

    def test_raise_typeerror(self) -> None:
        with pytest.raises(TypeError):
            parser = ImageFile.Parser()
            parser.feed(1)  # type: ignore[arg-type]

    def test_negative_stride(self) -> None:
        with open("Tests/images/raw_negative_stride.bin", "rb") as f:
            input = f.read()
        p = ImageFile.Parser()
        p.feed(input)
        with pytest.raises(OSError):
            p.close()

    def test_negative_offset(self) -> None:
        with Image.open("Tests/images/raw_negative_stride.bin") as im:
            with pytest.raises(ValueError, match="Tile offset cannot be negative"):
                im.load()

    @pytest.mark.parametrize("xy", ((-1, 0), (0, -1)))
    def test_negative_tile_extents(self, xy: tuple[int, int]) -> None:
        im = Image.new("1", (1, 1))
        fp = BytesIO()
        with pytest.raises(SystemError, match="tile cannot extend outside image"):
            ImageFile._save(im, fp, [ImageFile._Tile("raw", xy + (1, 1), 0, "1")])

    def test_extents_none(self) -> None:
        with Image.open("Tests/images/hopper.jpg") as im:
            im.tile = [im.tile[0]._replace(extents=None)]
            im.load()

        for extents in ("invalid", (0,), ("0", "0", "0", "0")):
            with Image.open("Tests/images/hopper.jpg") as im:
                im.tile = [im.tile[0]._replace(extents=extents)]  # type: ignore[arg-type]
                with pytest.raises(ValueError, match="invalid extents"):
                    im.load()

        im2 = Image.new("L", (1, 1))
        fp = BytesIO()
        tile = ImageFile._Tile("jpeg", None, 0, "L")
        ImageFile._save(im2, fp, [tile])

        for extents in ("invalid", (0,), ("0", "0", "0", "0")):
            tile = tile._replace(extents=extents)  # type: ignore[arg-type]
            with pytest.raises(ValueError, match="invalid extents"):
                ImageFile._save(im2, fp, [tile])

    def test_no_format(self) -> None:
        buf = BytesIO(b"\x00" * 255)

        class DummyImageFile(ImageFile.ImageFile):
            def _open(self) -> None:
                self._mode = "RGB"
                self._size = (1, 1)

        im = DummyImageFile(buf)
        assert im.format is None
        assert im.get_format_mimetype() is None

    def test_oserror(self) -> None:
        im = Image.new("RGB", (1, 1))
        with pytest.raises(OSError):
            im.save(BytesIO(), "JPEG2000", num_resolutions=2)

    def test_truncated(self) -> None:
        b = BytesIO(
            b"BM000000000000"  # head_data
            + _binary.o32le(
                ImageFile.SAFEBLOCK + 1 + 4
            )  # header_size, so BmpImagePlugin will try to read SAFEBLOCK + 1 bytes
            + (
                b"0" * ImageFile.SAFEBLOCK
            )  # only SAFEBLOCK bytes, so that the header is truncated
        )
        with pytest.raises(OSError, match="Truncated File Read"):
            BmpImagePlugin.BmpImageFile(b)

    @skip_unless_feature("zlib")
    def test_truncated_with_errors(self) -> None:
        with Image.open("Tests/images/truncated_image.png") as im:
            with pytest.raises(OSError):
                im.load()

            # Test that the error is raised if loaded a second time
            with pytest.raises(OSError):
                im.load()

    @skip_unless_feature("zlib")
    def test_truncated_without_errors(self, monkeypatch: pytest.MonkeyPatch) -> None:
        with Image.open("Tests/images/truncated_image.png") as im:
            monkeypatch.setattr(ImageFile, "LOAD_TRUNCATED_IMAGES", True)
            im.load()

    @skip_unless_feature("zlib")
    def test_broken_datastream_with_errors(self) -> None:
        with Image.open("Tests/images/broken_data_stream.png") as im:
            with pytest.raises(OSError):
                im.load()

    @skip_unless_feature("zlib")
    def test_broken_datastream_without_errors(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with Image.open("Tests/images/broken_data_stream.png") as im:
            monkeypatch.setattr(ImageFile, "LOAD_TRUNCATED_IMAGES", True)
            im.load()


class MockPyDecoder(ImageFile.PyDecoder):
    last: MockPyDecoder

    def __init__(self, mode: str, *args: Any) -> None:
        MockPyDecoder.last = self

        super().__init__(mode, *args)

    def decode(self, buffer: Image.DecoderInput) -> tuple[int, int]:
        # eof
        return -1, 0


class MockPyEncoder(ImageFile.PyEncoder):
    last: MockPyEncoder | None

    def __init__(self, mode: str, *args: Any) -> None:
        MockPyEncoder.last = self

        super().__init__(mode, *args)

    def encode(self, bufsize: int) -> tuple[int, int, bytes]:
        return 1, 1, b""

    def cleanup(self) -> None:
        self.cleanup_called = True


xoff, yoff, xsize, ysize = 10, 20, 100, 100


class MockImageFile(ImageFile.ImageFile):
    def _open(self) -> None:
        self.rawmode = "RGBA"
        self._mode = "RGBA"
        self._size = (200, 200)
        self.tile = [
            ImageFile._Tile("MOCK", (xoff, yoff, xoff + xsize, yoff + ysize), 32, None)
        ]


class CodecsTest:
    @classmethod
    def setup_class(cls) -> None:
        Image.register_decoder("MOCK", MockPyDecoder)
        Image.register_encoder("MOCK", MockPyEncoder)


class TestPyDecoder(CodecsTest):
    def test_setimage(self) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)

        im.load()

        assert MockPyDecoder.last.state.xoff == xoff
        assert MockPyDecoder.last.state.yoff == yoff
        assert MockPyDecoder.last.state.xsize == xsize
        assert MockPyDecoder.last.state.ysize == ysize

        with pytest.raises(ValueError):
            MockPyDecoder.last.set_as_raw(b"\x00")

    @pytest.mark.parametrize(
        "extents",
        (
            (-10, yoff, xoff + xsize, yoff + ysize),
            (xoff, -10, xoff + xsize, yoff + ysize),
            (xoff, yoff, -10, yoff + ysize),
            (xoff, yoff, xoff + xsize, -10),
            (xoff, yoff, xoff + xsize + 100, yoff + ysize),
            (xoff, yoff, xoff + xsize, yoff + ysize + 100),
        ),
    )
    def test_extents(self, extents: tuple[int, int, int, int]) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [ImageFile._Tile("MOCK", extents, 32, None)]

        with pytest.raises(ValueError):
            im.load()

    def test_extents_none(self) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [ImageFile._Tile("MOCK", None, 32, None)]

        im.load()

        assert MockPyDecoder.last.state.xoff == 0
        assert MockPyDecoder.last.state.yoff == 0
        assert MockPyDecoder.last.state.xsize == 200
        assert MockPyDecoder.last.state.ysize == 200

    def test_decode(self) -> None:
        decoder = ImageFile.PyDecoder("")
        with pytest.raises(NotImplementedError):
            decoder.decode(b"")


class TestPyEncoder(CodecsTest):
    def test_setimage(self) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)

        fp = BytesIO()
        ImageFile._save(
            im,
            fp,
            [
                ImageFile._Tile(
                    "MOCK", (xoff, yoff, xoff + xsize, yoff + ysize), 0, "RGB"
                )
            ],
        )

        assert MockPyEncoder.last
        assert MockPyEncoder.last.state.xoff == xoff
        assert MockPyEncoder.last.state.yoff == yoff
        assert MockPyEncoder.last.state.xsize == xsize
        assert MockPyEncoder.last.state.ysize == ysize

    @pytest.mark.parametrize(
        "extents",
        (
            (-10, yoff, xoff + xsize, yoff + ysize),
            (xoff, -10, xoff + xsize, yoff + ysize),
            (xoff, yoff, -10, yoff + ysize),
            (xoff, yoff, xoff + xsize, -10),
            (xoff, yoff, xoff + xsize + 100, yoff + ysize),
            (xoff, yoff, xoff + xsize, yoff + ysize + 100),
        ),
    )
    def test_extents(self, extents: tuple[int, int, int, int]) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)

        fp = BytesIO()
        MockPyEncoder.last = None
        with pytest.raises(ValueError):
            ImageFile._save(im, fp, [ImageFile._Tile("MOCK", extents, 0, "RGB")])
        last: MockPyEncoder | None = MockPyEncoder.last
        assert last
        assert last.cleanup_called

        with pytest.raises(ValueError):
            ImageFile._save(im, fp, [ImageFile._Tile("MOCK", extents, 0, "RGB")])

    def test_extents_none(self) -> None:
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [ImageFile._Tile("MOCK", None, 32, None)]

        fp = BytesIO()
        ImageFile._save(im, fp, [ImageFile._Tile("MOCK", None, 0, "RGB")])

        assert MockPyEncoder.last
        assert MockPyEncoder.last.state.xoff == 0
        assert MockPyEncoder.last.state.yoff == 0
        assert MockPyEncoder.last.state.xsize == 200
        assert MockPyEncoder.last.state.ysize == 200

    def test_encode(self) -> None:
        encoder = ImageFile.PyEncoder("")
        with pytest.raises(NotImplementedError):
            encoder.encode(0)

        bytes_consumed, errcode = encoder.encode_to_pyfd()
        assert bytes_consumed == 0
        assert ImageFile.ERRORS[errcode] == "bad configuration"

        encoder._pushes_fd = True
        with pytest.raises(NotImplementedError):
            encoder.encode_to_pyfd()

        with pytest.raises(NotImplementedError):
            encoder.encode_to_file(0, 0)

    def test_zero_height(self) -> None:
        with pytest.raises(UnidentifiedImageError):
            Image.open("Tests/images/zero_height.j2k")



from __future__ import annotations

import io
import logging
import os
import shutil
import sys
import tempfile
import warnings
from pathlib import Path
from types import ModuleType
from typing import IO, Any

import pytest

from PIL import (
    ExifTags,
    Image,
    ImageDraw,
    ImageFile,
    ImagePalette,
    ImageShow,
    UnidentifiedImageError,
    features,
)

from .helper import (
    assert_image_equal,
    assert_image_equal_tofile,
    assert_image_similar,
    assert_image_similar_tofile,
    assert_not_all_same,
    hopper,
    is_win32,
    mark_if_feature_version,
    skip_unless_feature,
    timeout_unless_slower_valgrind,
)

ElementTree: ModuleType | None
try:
    from defusedxml import ElementTree
except ImportError:
    ElementTree = None

PrettyPrinter: type | None
try:
    from IPython.lib.pretty import PrettyPrinter
except ImportError:
    PrettyPrinter = None


class TestImage:
    @pytest.mark.parametrize("mode", Image.MODES)
    def test_image_modes_success(self, mode: str) -> None:
        Image.new(mode, (1, 1))

    @pytest.mark.parametrize("mode", ("", "bad", "very very long"))
    def test_image_modes_fail(self, mode: str) -> None:
        with pytest.raises(ValueError, match="unrecognized image mode"):
            Image.new(mode, (1, 1))

    def test_exception_inheritance(self) -> None:
        assert issubclass(UnidentifiedImageError, OSError)

    def test_sanity(self) -> None:
        im = Image.new("L", (100, 100))
        assert repr(im).startswith("<PIL.Image.Image image mode=L size=100x100 at")
        assert im.mode == "L"
        assert im.size == (100, 100)

        im = Image.new("RGB", (100, 100))
        assert repr(im).startswith("<PIL.Image.Image image mode=RGB size=100x100 ")
        assert im.mode == "RGB"
        assert im.size == (100, 100)

        Image.new("L", (100, 100), None)
        im2 = Image.new("L", (100, 100), 0)
        im3 = Image.new("L", (100, 100), "black")

        assert im2.getcolors() == [(10000, 0)]
        assert im3.getcolors() == [(10000, 0)]

        with pytest.raises(ValueError):
            Image.new("X", (100, 100))
        with pytest.raises(ValueError):
            Image.new("", (100, 100))
        # with pytest.raises(MemoryError):
        #   Image.new("L", (1000000, 1000000))

    @pytest.mark.skipif(PrettyPrinter is None, reason="IPython is not installed")
    def test_repr_pretty(self) -> None:
        im = Image.new("L", (100, 100))

        output = io.StringIO()
        assert PrettyPrinter is not None
        p = PrettyPrinter(output)
        im._repr_pretty_(p, False)
        assert output.getvalue() == "<PIL.Image.Image image mode=L size=100x100>"

    def test_open_formats(self) -> None:
        PNGFILE = "Tests/images/hopper.png"
        JPGFILE = "Tests/images/hopper.jpg"

        with pytest.raises(TypeError):
            with Image.open(PNGFILE, formats=123):  # type: ignore[arg-type]
                pass

        format_list: list[list[str] | tuple[str, ...]] = [
            ["JPEG"],
            ("JPEG",),
            ["jpeg"],
            ["Jpeg"],
            ["jPeG"],
            ["JpEg"],
        ]
        for formats in format_list:
            with pytest.raises(UnidentifiedImageError):
                with Image.open(PNGFILE, formats=formats):
                    pass

            with Image.open(JPGFILE, formats=formats) as im:
                assert im.mode == "RGB"
                assert im.size == (128, 128)

        for file in [PNGFILE, JPGFILE]:
            with Image.open(file, formats=None) as im:
                assert im.mode == "RGB"
                assert im.size == (128, 128)

    def test_open_verbose_failure(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(Image, "WARN_POSSIBLE_FORMATS", True)

        im = io.BytesIO(b"")
        with pytest.raises(UnidentifiedImageError):
            with pytest.warns(UserWarning, match="opening failed"):
                with Image.open(im):
                    pass

    def test_width_height(self) -> None:
        im = Image.new("RGB", (1, 2))
        assert im.width == 1
        assert im.height == 2

        with pytest.raises(AttributeError):
            im.size = (3, 4)  # type: ignore[misc]

    def test_set_mode(self) -> None:
        im = Image.new("RGB", (1, 1))

        with pytest.raises(AttributeError):
            im.mode = "P"  # type: ignore[misc]

    def test_empty_path(self) -> None:
        with pytest.raises(FileNotFoundError):
            Image.open("")

    def test_invalid_image(self) -> None:
        im = io.BytesIO(b"")
        with pytest.raises(UnidentifiedImageError):
            with Image.open(im):
                pass

    def test_bad_mode(self) -> None:
        with pytest.raises(ValueError):
            with Image.open("filename", "bad mode"):  # type: ignore[arg-type]
                pass

    def test_stringio(self) -> None:
        with pytest.raises(ValueError):
            with Image.open(io.StringIO()):  # type: ignore[arg-type]
                pass

    def test_string(self, tmp_path: Path) -> None:
        out = str(tmp_path / "temp.png")
        im = hopper()
        im.save(out)
        with Image.open(out) as reloaded:
            assert_image_equal(im, reloaded)

    def test_pathlib(self, tmp_path: Path) -> None:
        with Image.open(Path("Tests/images/multipage-mmap.tiff")) as im:
            assert im.mode == "P"
            assert im.size == (10, 10)

        with Image.open(Path("Tests/images/hopper.jpg")) as im:
            assert im.mode == "RGB"
            assert im.size == (128, 128)

            for ext in (".jpg", ".jp2"):
                if ext == ".jp2" and not features.check_codec("jpg_2000"):
                    pytest.skip("jpg_2000 not available")
                im.save(tmp_path / ("temp." + ext))

    def test_fp_name(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.jpg"

        class FP(io.BytesIO):
            name: Path

            if sys.version_info >= (3, 12):
                from collections.abc import Buffer

                def write(self, data: Buffer) -> int:
                    return len(data)

            else:

                def write(self, data: Any) -> int:
                    return len(data)

        fp = FP()
        fp.name = temp_file

        im = hopper()
        im.save(fp)

    def test_tempfile(self) -> None:
        # see #1460, pathlib support breaks tempfile.TemporaryFile on py27
        # Will error out on save on 3.0.0
        im = hopper()
        with tempfile.TemporaryFile() as fp:
            im.save(fp, "JPEG")
            fp.seek(0)
            with Image.open(fp) as reloaded:
                assert_image_similar(im, reloaded, 20)

    def test_unknown_extension(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.unknown"
        with hopper() as im:
            with pytest.raises(ValueError):
                im.save(temp_file)

    def test_internals(self) -> None:
        im = Image.new("L", (100, 100))
        im.readonly = 1
        im._copy()
        assert not im.readonly

        im.readonly = 1
        im.paste(0, (0, 0, 100, 100))
        assert not im.readonly

    @pytest.mark.skipif(is_win32(), reason="Test requires opening tempfile twice")
    @pytest.mark.skipif(
        sys.platform == "cygwin",
        reason="Test requires opening an mmaped file for writing",
    )
    def test_readonly_save(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.bmp"
        shutil.copy("Tests/images/rgb32bf-rgba.bmp", temp_file)

        with Image.open(temp_file) as im:
            assert im.readonly
            im.save(temp_file)

    def test_save_without_changing_readonly(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.bmp"

        with Image.open("Tests/images/rgb32bf-rgba.bmp") as im:
            assert im.readonly

            im.save(temp_file)
            assert im.readonly

    def test_dump(self, tmp_path: Path) -> None:
        im = Image.new("L", (10, 10))
        im._dump(str(tmp_path / "temp_L.ppm"))

        im = Image.new("RGB", (10, 10))
        im._dump(str(tmp_path / "temp_RGB.ppm"))

        im = Image.new("RGBA", (10, 10))
        im._dump(str(tmp_path / "temp_RGBA.ppm"))

        im = Image.new("HSV", (10, 10))
        with pytest.raises(OSError):
            im._dump(str(tmp_path / "temp_HSV.ppm"))

    def test_comparison_with_other_type(self) -> None:
        # Arrange
        item = Image.new("RGB", (25, 25), "#000")
        num = 12

        # Act/Assert
        # Shouldn't cause AttributeError (#774)
        assert item is not None
        assert item != num

    def test_getbands(self) -> None:
        # Assert
        assert hopper("RGB").getbands() == ("R", "G", "B")
        assert hopper("YCbCr").getbands() == ("Y", "Cb", "Cr")

    def test_getchannel_wrong_params(self) -> None:
        im = hopper()

        with pytest.raises(ValueError):
            im.getchannel(-1)
        with pytest.raises(ValueError):
            im.getchannel(3)
        with pytest.raises(ValueError):
            im.getchannel("Z")
        with pytest.raises(ValueError):
            im.getchannel("1")

    def test_getchannel(self) -> None:
        im = hopper("YCbCr")
        Y, Cb, Cr = im.split()

        assert_image_equal(Y, im.getchannel(0))
        assert_image_equal(Y, im.getchannel("Y"))
        assert_image_equal(Cb, im.getchannel(1))
        assert_image_equal(Cb, im.getchannel("Cb"))
        assert_image_equal(Cr, im.getchannel(2))
        assert_image_equal(Cr, im.getchannel("Cr"))

    def test_getbbox(self) -> None:
        # Arrange
        im = hopper()

        # Act
        bbox = im.getbbox()

        # Assert
        assert bbox == (0, 0, 128, 128)

    def test_ne(self) -> None:
        # Arrange
        im1 = Image.new("RGB", (25, 25), "black")
        im2 = Image.new("RGB", (25, 25), "white")

        # Act / Assert
        assert im1 != im2

    def test_alpha_composite(self) -> None:
        # https://stackoverflow.com/questions/3374878
        # Arrange
        expected_colors = sorted(
            [
                (1122, (128, 127, 0, 255)),
                (1089, (0, 255, 0, 255)),
                (3300, (255, 0, 0, 255)),
                (1156, (170, 85, 0, 192)),
                (1122, (0, 255, 0, 128)),
                (1122, (255, 0, 0, 128)),
                (1089, (0, 255, 0, 0)),
            ]
        )

        dst = Image.new("RGBA", size=(100, 100), color=(0, 255, 0, 255))
        draw = ImageDraw.Draw(dst)
        draw.rectangle((0, 33, 100, 66), fill=(0, 255, 0, 128))
        draw.rectangle((0, 67, 100, 100), fill=(0, 255, 0, 0))
        src = Image.new("RGBA", size=(100, 100), color=(255, 0, 0, 255))
        draw = ImageDraw.Draw(src)
        draw.rectangle((33, 0, 66, 100), fill=(255, 0, 0, 128))
        draw.rectangle((67, 0, 100, 100), fill=(255, 0, 0, 0))

        # Act
        img = Image.alpha_composite(dst, src)

        # Assert
        img_colors = img.getcolors()
        assert img_colors is not None
        assert sorted(img_colors) == expected_colors

    def test_alpha_composite_la(self) -> None:
        # Arrange
        expected_colors = sorted(
            [
                (3300, (255, 255)),
                (1156, (170, 192)),
                (1122, (128, 255)),
                (1089, (0, 0)),
                (1122, (255, 128)),
                (1122, (0, 128)),
                (1089, (0, 255)),
            ]
        )

        dst = Image.new("LA", size=(100, 100), color=(0, 255))
        draw = ImageDraw.Draw(dst)
        draw.rectangle((0, 33, 100, 66), fill=(0, 128))
        draw.rectangle((0, 67, 100, 100), fill=(0, 0))
        src = Image.new("LA", size=(100, 100), color=(255, 255))
        draw = ImageDraw.Draw(src)
        draw.rectangle((33, 0, 66, 100), fill=(255, 128))
        draw.rectangle((67, 0, 100, 100), fill=(255, 0))

        # Act
        img = Image.alpha_composite(dst, src)

        # Assert
        img_colors = img.getcolors()
        assert img_colors is not None
        assert sorted(img_colors) == expected_colors

    def test_alpha_inplace(self) -> None:
        src = Image.new("RGBA", (128, 128), "blue")

        over = Image.new("RGBA", (128, 128), "red")
        mask = hopper("L")
        over.putalpha(mask)

        target = Image.alpha_composite(src, over)

        # basic
        full = src.copy()
        full.alpha_composite(over)
        assert_image_equal(full, target)

        # with offset down to right
        offset = src.copy()
        offset.alpha_composite(over, (64, 64))
        assert_image_equal(offset.crop((64, 64, 127, 127)), target.crop((0, 0, 63, 63)))
        assert offset.size == (128, 128)

        # with negative offset
        offset = src.copy()
        offset.alpha_composite(over, (-64, -64))
        assert_image_equal(offset.crop((0, 0, 63, 63)), target.crop((64, 64, 127, 127)))
        assert offset.size == (128, 128)

        # offset and crop
        box = src.copy()
        box.alpha_composite(over, (64, 64), (0, 0, 32, 32))
        assert_image_equal(box.crop((64, 64, 96, 96)), target.crop((0, 0, 32, 32)))
        assert_image_equal(box.crop((96, 96, 128, 128)), src.crop((0, 0, 32, 32)))
        assert box.size == (128, 128)

        # source point
        source = src.copy()
        source.alpha_composite(over, (32, 32), (32, 32, 96, 96))

        assert_image_equal(source.crop((32, 32, 96, 96)), target.crop((32, 32, 96, 96)))
        assert source.size == (128, 128)

        # errors
        with pytest.raises(ValueError):
            source.alpha_composite(over, "invalid destination")  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            source.alpha_composite(over, (0, 0), "invalid source")  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            source.alpha_composite(over, 0)  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            source.alpha_composite(over, (0, 0), 0)  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            source.alpha_composite(over, (0, 0), (0, -1))

    def test_register_open_duplicates(self) -> None:
        # Arrange
        factory, accept = Image.OPEN["JPEG"]
        id_length = len(Image.ID)

        # Act
        Image.register_open("JPEG", factory, accept)

        # Assert
        assert len(Image.ID) == id_length

    def test_registered_extensions_uninitialized(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        # Arrange
        monkeypatch.setattr(Image, "_initialized", 0)

        # Act
        Image.registered_extensions()

        # Assert
        assert Image._initialized == 2

        for extension in Image.EXTENSION:
            assert extension in Image._EXTENSION_PLUGIN

    def test_registered_extensions(self) -> None:
        # Arrange
        # Open an image to trigger plugin registration
        with Image.open("Tests/images/rgb.jpg"):
            pass

        # Act
        extensions = Image.registered_extensions()

        # Assert
        assert extensions
        for ext in [".cur", ".icns", ".tif", ".tiff"]:
            assert ext in extensions

    def test_effect_mandelbrot(self) -> None:
        # Arrange
        size = (512, 512)
        extent = (-3, -2.5, 2, 2.5)
        quality = 100

        # Act
        im = Image.effect_mandelbrot(size, extent, quality)

        # Assert
        assert im.size == (512, 512)
        assert_image_equal_tofile(im, "Tests/images/effect_mandelbrot.png")

    def test_effect_mandelbrot_bad_arguments(self) -> None:
        # Arrange
        size = (512, 512)
        # Get coordinates the wrong way round:
        extent = (+3, +2.5, -2, -2.5)
        # Quality < 2:
        quality = 1

        # Act/Assert
        with pytest.raises(ValueError):
            Image.effect_mandelbrot(size, extent, quality)

    def test_effect_noise(self) -> None:
        # Arrange
        size = (100, 100)
        sigma = 128

        # Act
        im = Image.effect_noise(size, sigma)

        # Assert
        assert im.size == (100, 100)
        assert im.mode == "L"
        p0 = im.getpixel((0, 0))
        p1 = im.getpixel((0, 1))
        p2 = im.getpixel((0, 2))
        p3 = im.getpixel((0, 3))
        p4 = im.getpixel((0, 4))
        assert_not_all_same([p0, p1, p2, p3, p4])

    def test_effect_spread(self) -> None:
        # Arrange
        im = hopper()
        distance = 10

        # Act
        im2 = im.effect_spread(distance)

        # Assert
        assert im.size == (128, 128)
        assert_image_similar_tofile(im2, "Tests/images/effect_spread.png", 110)

    def test_effect_spread_zero(self) -> None:
        # Arrange
        im = hopper()
        distance = 0

        # Act
        im2 = im.effect_spread(distance)

        # Assert
        assert_image_equal(im, im2)

    def test_check_size(self) -> None:
        # Checking that the _check_size function throws value errors when we want it to
        with pytest.raises(ValueError):
            # not a tuple
            Image.new("RGB", 0)  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            # tuple too short
            Image.new("RGB", (0,))  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            Image.new("RGB", (-1, -1))  # w,h < 0

        # this should pass with 0 sized images, #2259
        im = Image.new("L", (0, 0))
        assert im.size == (0, 0)

        im = Image.new("L", (0, 100))
        assert im.size == (0, 100)

        im = Image.new("L", (100, 0))
        assert im.size == (100, 0)

        assert Image.new("RGB", (1, 1))
        # Should pass lists too
        i = Image.new("RGB", [1, 1])
        assert isinstance(i.size, tuple)

    @timeout_unless_slower_valgrind(0.75)
    @pytest.mark.parametrize("size", ((0, 100000000), (100000000, 0)))
    def test_empty_image(self, size: tuple[int, int]) -> None:
        Image.new("RGB", size)

    def test_storage_neg(self) -> None:
        # Storage.c accepted negative values for xsize, ysize.  Was
        # test_neg_ppm, but the core function for that has been
        # removed Calling directly into core to test the error in
        # Storage.c, rather than the size check above

        with pytest.raises(ValueError):
            Image.core.fill("RGB", (2, -2), (0, 0, 0))

    def test_one_item_tuple(self) -> None:
        for mode in ("I", "F", "L"):
            im = Image.new(mode, (100, 100), (5,))
            assert im.getpixel((0, 0)) == 5

    def test_linear_gradient_wrong_mode(self) -> None:
        # Arrange
        wrong_mode = "RGB"

        # Act / Assert
        with pytest.raises(ValueError):
            Image.linear_gradient(wrong_mode)

    @pytest.mark.parametrize("mode", ("L", "P", "I", "F"))
    def test_linear_gradient(self, mode: str) -> None:
        # Arrange
        target_file = "Tests/images/linear_gradient.png"

        # Act
        im = Image.linear_gradient(mode)

        # Assert
        assert im.size == (256, 256)
        assert im.mode == mode
        assert im.getpixel((0, 0)) == 0
        assert im.getpixel((255, 255)) == 255
        with Image.open(target_file) as target:
            im_target = target.convert(mode)
        assert_image_equal(im, im_target)

    def test_radial_gradient_wrong_mode(self) -> None:
        # Arrange
        wrong_mode = "RGB"

        # Act / Assert
        with pytest.raises(ValueError):
            Image.radial_gradient(wrong_mode)

    @pytest.mark.parametrize("mode", ("L", "P", "I", "F"))
    def test_radial_gradient(self, mode: str) -> None:
        # Arrange
        target_file = "Tests/images/radial_gradient.png"

        # Act
        im = Image.radial_gradient(mode)

        # Assert
        assert im.size == (256, 256)
        assert im.mode == mode
        assert im.getpixel((0, 0)) == 255
        assert im.getpixel((128, 128)) == 0
        with Image.open(target_file) as target:
            im_target = target.convert(mode)
        assert_image_equal(im, im_target)

    def test_register_extensions(self) -> None:
        test_format = "a"
        exts = ["b", "c"]
        for ext in exts:
            Image.register_extension(test_format, ext)
        ext_individual = Image.EXTENSION.copy()
        for ext in exts:
            del Image.EXTENSION[ext]

        Image.register_extensions(test_format, exts)
        ext_multiple = Image.EXTENSION.copy()
        for ext in exts:
            del Image.EXTENSION[ext]

        assert ext_individual == ext_multiple

    def test_remap_palette(self) -> None:
        # Test identity transform
        with Image.open("Tests/images/hopper.gif") as im:
            assert_image_equal(im, im.remap_palette(list(range(256))))

        # Test identity transform with an RGBA palette
        im_p = Image.new("P", (256, 1))
        for x in range(256):
            im_p.putpixel((x, 0), x)
        im_p.putpalette(list(range(256)) * 4, "RGBA")
        im_remapped = im_p.remap_palette(list(range(256)))
        assert_image_equal(im_p, im_remapped)
        assert im_p.palette is not None
        assert im_remapped.palette is not None
        assert im_p.palette.palette == im_remapped.palette.palette

        # Test illegal image mode
        with hopper() as im_hopper:
            with pytest.raises(ValueError):
                im_hopper.remap_palette([])

    def test_remap_palette_transparency(self) -> None:
        im = Image.new("P", (1, 2), (0, 0, 0))
        im.putpixel((0, 1), (255, 0, 0))
        im.info["transparency"] = 0

        im_remapped = im.remap_palette([1, 0])
        assert im_remapped.info["transparency"] == 1
        palette = im_remapped.getpalette()
        assert palette is not None
        assert len(palette) == 6

        # Test unused transparency
        im.info["transparency"] = 2

        im_remapped = im.remap_palette([1, 0])
        assert "transparency" not in im_remapped.info

    def test__new(self) -> None:
        im = hopper("RGB")
        im_p = hopper("P")

        blank_p = Image.new("P", (10, 10))
        blank_pa = Image.new("PA", (10, 10))
        blank_p.palette = None
        blank_pa.palette = None

        def _make_new(
            base_image: Image.Image,
            image: Image.Image,
            palette_result: ImagePalette.ImagePalette | None = None,
        ) -> None:
            new_image = base_image._new(image.im)
            assert new_image.mode == image.mode
            assert new_image.size == image.size
            assert new_image.info == base_image.info
            if palette_result is not None:
                assert new_image.palette is not None
                assert new_image.palette.tobytes() == palette_result.tobytes()
            else:
                assert new_image.palette is None

        _make_new(im, im_p, ImagePalette.ImagePalette("RGB"))
        _make_new(im_p, im, None)
        _make_new(im, blank_p, ImagePalette.ImagePalette())
        _make_new(im, blank_pa, ImagePalette.ImagePalette())

    @pytest.mark.parametrize(
        "mode, color",
        (
            ("RGB", "#DDEEFF"),
            ("RGB", (221, 238, 255)),
            ("RGBA", (221, 238, 255, 255)),
        ),
    )
    def test_p_from_rgb_rgba(self, mode: str, color: str | tuple[int, ...]) -> None:
        im = Image.new("P", (100, 100), color)
        expected = Image.new(mode, (100, 100), color)
        assert_image_equal(im.convert(mode), expected)

    def test_no_resource_warning_on_save(self, tmp_path: Path) -> None:
        # https://github.com/python-pillow/Pillow/issues/835
        # Arrange
        test_file = "Tests/images/hopper.png"
        temp_file = tmp_path / "temp.jpg"

        # Act/Assert
        with Image.open(test_file) as im:
            with warnings.catch_warnings():
                warnings.simplefilter("error")

                im.save(temp_file)

    def test_no_new_file_on_error(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.jpg"

        im = Image.new("RGB", (0, 0))
        with pytest.raises(ValueError):
            im.save(temp_file)

        assert not os.path.exists(temp_file)

    def test_load_on_nonexclusive_multiframe(self) -> None:
        with open("Tests/images/frozenpond.mpo", "rb") as fp:

            def act(fp: IO[bytes]) -> None:
                im = Image.open(fp)
                im.load()

            act(fp)

            with Image.open(fp) as im:
                im.load()

            assert not fp.closed

    def test_empty_exif(self) -> None:
        with Image.open("Tests/images/exif.png") as im:
            exif = im.getexif()
        assert dict(exif)

        # Test that exif data is cleared after another load
        exif.load(b"")
        assert not dict(exif)

        # Test loading just the EXIF header
        exif.load(b"Exif\x00\x00")
        assert not dict(exif)

    def test_duplicate_exif_header(self) -> None:
        with Image.open("Tests/images/exif.png") as im:
            im.load()
            im.info["exif"] = b"Exif\x00\x00" + im.info["exif"]

            exif = im.getexif()
        assert exif[274] == 1

    def test_empty_get_ifd(self) -> None:
        exif = Image.Exif()
        ifd = exif.get_ifd(0x8769)
        assert ifd == {}

        ifd[36864] = b"0220"
        assert exif.get_ifd(0x8769) == {36864: b"0220"}

        reloaded_exif = Image.Exif()
        reloaded_exif.load(exif.tobytes())
        assert reloaded_exif.get_ifd(0x8769) == {36864: b"0220"}

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_exif_jpeg(self, tmp_path: Path) -> None:
        with Image.open("Tests/images/exif-72dpi-int.jpg") as im:  # Little endian
            exif = im.getexif()
            assert 258 not in exif
            assert 274 in exif
            assert 282 in exif
            assert exif[296] == 2
            assert exif[11] == "gThumb 3.0.1"

            out = tmp_path / "temp.jpg"
            exif[258] = 8
            del exif[274]
            del exif[282]
            exif[296] = 455
            exif[11] = "Pillow test"
            im.save(out, exif=exif)
        with Image.open(out) as reloaded:
            reloaded_exif = reloaded.getexif()
            assert reloaded_exif[258] == 8
            assert 274 not in reloaded_exif
            assert 282 not in reloaded_exif
            assert reloaded_exif[296] == 455
            assert reloaded_exif[11] == "Pillow test"

        with Image.open("Tests/images/no-dpi-in-exif.jpg") as im:  # Big endian
            exif = im.getexif()
            assert 258 not in exif
            assert 306 in exif
            assert exif[274] == 1
            assert exif[305] == "Adobe Photoshop CC 2017 (Macintosh)"

            out = tmp_path / "temp.jpg"
            exif[258] = 8
            del exif[306]
            exif[274] = 455
            exif[305] = "Pillow test"
            im.save(out, exif=exif)
        with Image.open(out) as reloaded:
            reloaded_exif = reloaded.getexif()
            assert reloaded_exif[258] == 8
            assert 306 not in reloaded_exif
            assert reloaded_exif[274] == 455
            assert reloaded_exif[305] == "Pillow test"

    @skip_unless_feature("webp")
    def test_exif_webp(self, tmp_path: Path) -> None:
        with Image.open("Tests/images/hopper.webp") as im:
            exif = im.getexif()
            assert dict(exif) == {}

            out = tmp_path / "temp.webp"
            exif[258] = 8
            exif[40963] = 455
            exif[305] = "Pillow test"

            def check_exif() -> None:
                with Image.open(out) as reloaded:
                    reloaded_exif = reloaded.getexif()
                    assert reloaded_exif[258] == 8
                    assert reloaded_exif[40963] == 455
                    assert reloaded_exif[305] == "Pillow test"

            im.save(out, exif=exif)
            check_exif()
            im.save(out, exif=exif, save_all=True)
            check_exif()

    def test_exif_png(self, tmp_path: Path) -> None:
        with Image.open("Tests/images/exif.png") as im:
            exif = im.getexif()
            assert dict(exif) == {274: 1}

            out = tmp_path / "temp.png"
            exif[258] = 8
            del exif[274]
            exif[40963] = 455
            exif[305] = "Pillow test"
            im.save(out, exif=exif)

        with Image.open(out) as reloaded:
            reloaded_exif = reloaded.getexif()
            assert reloaded_exif == {258: 8, 40963: 455, 305: "Pillow test"}

    def test_exif_interop(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()
            assert exif.get_ifd(0xA005) == {
                1: "R98",
                2: b"0100",
                4097: 2272,
                4098: 1704,
            }

            reloaded_exif = Image.Exif()
            reloaded_exif.load(exif.tobytes())
            assert reloaded_exif.get_ifd(0xA005) == exif.get_ifd(0xA005)

    def test_exif_ifd1(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()
            assert exif.get_ifd(ExifTags.IFD.IFD1) == {
                513: 2036,
                514: 5448,
                259: 6,
                296: 2,
                282: 180.0,
                283: 180.0,
            }

    def test_exif_ifd(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()
        del exif.get_ifd(0x8769)[0xA005]

        reloaded_exif = Image.Exif()
        reloaded_exif.load(exif.tobytes())
        assert reloaded_exif.get_ifd(0x8769) == exif.get_ifd(0x8769)

    def test_delete_ifd_tag(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()
        exif.get_ifd(0x8769)
        assert 0x8769 in exif
        del exif[0x8769]

        reloaded_exif = Image.Exif()
        reloaded_exif.load(exif.tobytes())
        assert 0x8769 not in reloaded_exif

    def test_exif_load_from_fp(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            data = im.info["exif"]
            if data.startswith(b"Exif\x00\x00"):
                data = data[6:]
            fp = io.BytesIO(data)

            exif = Image.Exif()
            exif.load_from_fp(fp)
            assert exif == {
                271: "Canon",
                272: "Canon PowerShot S40",
                274: 1,
                282: 180.0,
                283: 180.0,
                296: 2,
                306: "2003:12:14 12:01:44",
                531: 1,
                34665: 196,
            }

    def test_exif_hide_offsets(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()

        # Check offsets are present initially
        assert 0x8769 in exif
        for tag in (0xA005, 0x927C):
            assert tag in exif.get_ifd(0x8769)
        assert exif.get_ifd(0xA005)
        loaded_exif = exif

        with Image.open("Tests/images/flower.jpg") as im:
            new_exif = im.getexif()

            for exif in (loaded_exif, new_exif):
                exif.hide_offsets()

                # Assert they are hidden afterwards,
                # but that the IFDs are still available
                assert 0x8769 not in exif
                assert exif.get_ifd(0x8769)
                for tag in (0xA005, 0x927C):
                    assert tag not in exif.get_ifd(0x8769)
                assert exif.get_ifd(0xA005)

    def test_exif_from_xmp_bytes(self) -> None:
        im = Image.new("RGB", (1, 1))
        im.info["xmp"] = b'\xff tiff:Orientation="2"'
        assert im.getexif()[274] == 2

    def test_empty_xmp(self) -> None:
        with Image.open("Tests/images/hopper.gif") as im:
            if ElementTree is None:
                with pytest.warns(
                    UserWarning,
                    match="XMP data cannot be read without defusedxml dependency",
                ):
                    xmp = im.getxmp()
            else:
                xmp = im.getxmp()
            assert xmp == {}

    def test_getxmp_padded(self) -> None:
        im = Image.new("RGB", (1, 1))
        im.info["xmp"] = (
            b'<?xpacket begin="\xef\xbb\xbf" id="W5M0MpCehiHzreSzNTczkc9d"?>\n'
            b'<x:xmpmeta xmlns:x="adobe:ns:meta/" />\n<?xpacket end="w"?>\x00\x00 '
        )
        if ElementTree is None:
            with pytest.warns(
                UserWarning,
                match="XMP data cannot be read without defusedxml dependency",
            ):
                assert im.getxmp() == {}
        else:
            assert im.getxmp() == {"xmpmeta": None}

    def test_get_child_images(self) -> None:
        im = Image.new("RGB", (1, 1))
        with pytest.warns(DeprecationWarning, match="Image.Image.get_child_images"):
            assert im.get_child_images() == []

    def test_show(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(ImageShow, "_viewers", [])

        im = Image.new("RGB", (1, 1))
        with pytest.warns(DeprecationWarning, match="Image._show"):
            Image._show(im)

    @pytest.mark.parametrize("size", ((1, 0), (0, 1), (0, 0)))
    def test_zero_tobytes(self, size: tuple[int, int]) -> None:
        im = Image.new("RGB", size)
        assert im.tobytes() == b""

    @pytest.mark.parametrize("size", ((1, 0), (0, 1), (0, 0)))
    def test_zero_frombytes(self, size: tuple[int, int]) -> None:
        Image.frombytes("RGB", size, b"")

        im = Image.new("RGB", size)
        im.frombytes(b"")

    def test_has_transparency_data(self) -> None:
        for mode in ("1", "L", "P", "RGB"):
            im = Image.new(mode, (1, 1))
            assert not im.has_transparency_data

        for mode in ("LA", "La", "PA", "RGBA", "RGBa"):
            im = Image.new(mode, (1, 1))
            assert im.has_transparency_data

        # P mode with "transparency" info
        with Image.open("Tests/images/first_frame_transparency.gif") as im:
            assert "transparency" in im.info
            assert im.has_transparency_data

        # RGB mode with "transparency" info
        with Image.open("Tests/images/rgb_trns.png") as im:
            assert "transparency" in im.info
            assert im.has_transparency_data

        # P mode with RGBA palette
        im = Image.new("RGBA", (1, 1)).convert("P")
        assert im.mode == "P"
        assert im.palette is not None
        assert im.palette.mode == "RGBA"
        assert im.has_transparency_data

    def test_apply_transparency(self) -> None:
        im = Image.new("P", (1, 1))
        im.putpalette((0, 0, 0, 1, 1, 1))
        assert im.palette is not None
        assert im.palette.colors == {(0, 0, 0): 0, (1, 1, 1): 1}

        # Test that no transformation is applied without transparency
        im.apply_transparency()
        assert im.palette.colors == {(0, 0, 0): 0, (1, 1, 1): 1}

        # Test that a transparency index is applied
        im.info["transparency"] = 0
        im.apply_transparency()
        assert "transparency" not in im.info
        assert im.palette.colors == {(0, 0, 0, 0): 0, (1, 1, 1, 255): 1}

        # Test that existing transparency is kept
        im = Image.new("P", (1, 1))
        im.putpalette((0, 0, 0, 255, 1, 1, 1, 128), "RGBA")
        im.info["transparency"] = 0
        im.apply_transparency()
        assert im.palette is not None
        assert im.palette.colors == {(0, 0, 0, 0): 0, (1, 1, 1, 128): 1}

        # Test that transparency bytes are applied
        with Image.open("Tests/images/pil123p.png") as im:
            assert isinstance(im.info["transparency"], bytes)
            assert im.palette is not None
            assert im.palette.colors[(27, 35, 6)] == 24
            im.apply_transparency()
            assert im.palette is not None
            assert im.palette.colors[(27, 35, 6, 214)] == 24

    def test_merge_pa(self) -> None:
        p = hopper("P")
        a = Image.new("L", p.size)
        pa = Image.merge("PA", (p, a))
        assert p.getpalette() == pa.getpalette()

    def test_constants(self) -> None:
        for enum in (
            Image.Transpose,
            Image.Transform,
            Image.Resampling,
            Image.Dither,
            Image.Palette,
            Image.Quantize,
        ):
            for name in enum.__members__:
                assert getattr(Image, name) == enum[name]

    def test_decoder_setimage_once(self) -> None:
        im = Image.new("L", (1, 1))
        decoder = Image._getdecoder("L", "raw", "L")

        decoder.setimage(im.im, None)
        with pytest.raises(ValueError, match="decoder already has an image"):
            decoder.setimage(im.im, None)

    @pytest.mark.parametrize(
        "path",
        [
            "fli_overrun.bin",
            "sgi_overrun.bin",
            "sgi_overrun_expandrow.bin",
            "sgi_overrun_expandrow2.bin",
            "pcx_overrun.bin",
            "pcx_overrun2.bin",
            "ossfuzz-4836216264589312.pcx",
            "01r_00.pcx",
        ],
    )
    def test_overrun(self, path: str) -> None:
        """For overrun completeness, test as:
        valgrind pytest -qq Tests/test_image.py::TestImage::test_overrun | grep decode.c
        """
        with Image.open(os.path.join("Tests/images", path)) as im:
            with pytest.raises(OSError) as e:
                im.load()
        buffer_overrun = str(e.value) == "buffer overrun when reading image file"
        truncated = "image file is truncated" in str(e.value)

        assert buffer_overrun or truncated

    def test_fli_overrun2(self) -> None:
        with Image.open("Tests/images/fli_overrun2.bin") as im:
            with pytest.raises(OSError, match="buffer overrun when reading image file"):
                im.seek(1)

    def test_exit_fp(self) -> None:
        with Image.new("L", (1, 1)) as im:
            pass
        assert not hasattr(im, "fp")

    def test_close_graceful(self, caplog: pytest.LogCaptureFixture) -> None:
        with Image.open("Tests/images/hopper.jpg") as im:
            copy = im.copy()
            with caplog.at_level(logging.DEBUG):
                im.close()
                copy.close()
            assert len(caplog.records) == 0
            assert im.fp is None


class TestImageBytes:
    @pytest.mark.parametrize("mode", Image.MODES)
    def test_roundtrip_bytes_constructor(self, mode: str) -> None:
        im = hopper(mode)
        source_bytes = im.tobytes()

        reloaded = Image.frombytes(mode, im.size, source_bytes)
        assert reloaded.tobytes() == source_bytes

    @pytest.mark.parametrize("mode", Image.MODES)
    def test_roundtrip_bytes_method(self, mode: str) -> None:
        im = hopper(mode)
        source_bytes = im.tobytes()

        reloaded = Image.new(mode, im.size)
        reloaded.frombytes(source_bytes)
        assert reloaded.tobytes() == source_bytes

    @pytest.mark.parametrize("mode", Image.MODES)
    def test_get_flattened_data_putdata(self, mode: str) -> None:
        im = hopper(mode)
        reloaded = Image.new(mode, im.size)
        reloaded.putdata(im.get_flattened_data())
        assert_image_equal(im, reloaded)


class MockEncoder(ImageFile.PyEncoder):
    pass


class TestRegistry:
    def test_encode_registry(self) -> None:
        Image.register_encoder("MOCK", MockEncoder)
        assert "MOCK" in Image.ENCODERS

        enc = Image._getencoder("RGB", "MOCK", ("args",), extra=("extra",))

        assert isinstance(enc, MockEncoder)
        assert enc.mode == "RGB"
        assert enc.args == ("args", "extra")

    def test_encode_registry_fail(self) -> None:
        with pytest.raises(OSError):
            Image._getencoder("RGB", "DoesNotExist", ("args",), extra=("extra",))



from __future__ import annotations

import io
import sys
import sysconfig

import pytest

FREE_THREADED_BUILD = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))

gil_enabled_at_start = True
if FREE_THREADED_BUILD:
    gil_enabled_at_start = sys._is_gil_enabled()  # type: ignore[attr-defined]


def pytest_report_header(config: pytest.Config) -> str:
    try:
        from PIL import features

        with io.StringIO() as out:
            features.pilinfo(out=out, supported_formats=False)
            return out.getvalue()
    except Exception as e:
        return f"pytest_report_header failed: {e}"


def pytest_terminal_summary(terminalreporter: pytest.TerminalReporter) -> None:
    if (
        FREE_THREADED_BUILD
        and not gil_enabled_at_start
        and sys._is_gil_enabled()  # type: ignore[attr-defined]
    ):
        tr = terminalreporter
        tr.ensure_newline()
        tr.section("GIL re-enabled", red=True, bold=True)
        tr.line("The GIL was re-enabled at runtime during the tests.")
        tr.line("This can happen with no test failures if the RuntimeWarning")
        tr.line("raised by Python when this happens is filtered by a test.")
        tr.line("")
        tr.line("Please ensure all new C modules declare support for running")
        tr.line("without the GIL. Any new tests that intentionally imports")
        tr.line("code that re-enables the GIL should do so in a subprocess.")
        pytest.exit("GIL re-enabled during tests", returncode=1)


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        "markers",
        "pil_noop_mark: A conditional mark where nothing special happens",
    )

    # We're marking some tests to ignore valgrind errors and XFAIL them.
    # Ensure that the mark is defined
    # even in cases where pytest-valgrind isn't installed
    try:
        config.addinivalue_line(
            "markers",
            "valgrind_known_error: Tests that have known issues with valgrind",
        )
    except Exception:
        # valgrind is already installed
        pass



from __future__ import annotations

import sys
from collections.abc import Callable
from typing import Any

import pytest

from PIL import Image

min_iterations = 100
max_iterations = 10000

pytestmark = pytest.mark.skipif(
    sys.platform.startswith("win32"), reason="requires Unix or macOS"
)


def _get_mem_usage() -> float:
    from resource import RUSAGE_SELF, getpagesize, getrusage

    mem = getrusage(RUSAGE_SELF).ru_maxrss
    return mem * getpagesize() / 1024 / 1024


def _test_leak(
    min_iterations: int,
    max_iterations: int,
    fn: Callable[..., Image.Image | None],
    *args: Any,
) -> None:
    mem_limit = None
    for i in range(max_iterations):
        fn(*args)
        mem = _get_mem_usage()
        if i < min_iterations:
            mem_limit = mem + 1
            continue
        msg = f"memory usage limit exceeded after {i + 1} iterations"
        assert mem_limit is not None
        assert mem <= mem_limit, msg


def test_leak_putdata() -> None:
    im = Image.new("RGB", (25, 25))
    _test_leak(min_iterations, max_iterations, im.putdata, im.getdata())


def test_leak_getlist() -> None:
    im = Image.new("P", (25, 25))
    _test_leak(
        min_iterations,
        max_iterations,
        # Pass a new list at each iteration.
        lambda: im.point(range(256)),
    )



from __future__ import annotations

import pytest

from .helper import assert_image_equal, fromstring, hopper


def test_sanity() -> None:
    with pytest.raises(ValueError):
        hopper().tobitmap()

    im1 = hopper().convert("1")

    bitmap = im1.tobitmap()

    assert isinstance(bitmap, bytes)
    assert_image_equal(im1, fromstring(bitmap))



from __future__ import annotations

from io import BytesIO
from pathlib import Path

import pytest

from PIL import FontFile, Image


def test_puti16() -> None:
    fp = BytesIO()
    FontFile.puti16(fp, (0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
    assert fp.getvalue() == (
        b"\x00\x00\x00\x01\x00\x02\x00\x03\x00\x04"
        b"\x00\x05\x00\x06\x00\x07\x00\x08\x00\t"
    )


def test_compile() -> None:
    font = FontFile.FontFile()
    font.glyph[0] = ((0, 0), (0, 0, 0, 0), (0, 0, 0, 1), Image.new("L", (0, 0)))
    font.compile()
    assert font.ysize == 1

    font.ysize = 2
    font.compile()

    # Assert that compiling again did not change anything
    assert font.ysize == 2


def test_decompression_bomb() -> None:
    font = FontFile.FontFile()
    font.glyph[0] = ((0, 0), (0, 0, 0, 0), (0, 0, 10000, 10000), Image.new("L", (0, 0)))
    with pytest.raises(Image.DecompressionBombError):
        font.compile()


def test_save(tmp_path: Path) -> None:
    tempname = str(tmp_path / "temp.pil")

    font = FontFile.FontFile()
    with pytest.raises(ValueError, match="No bitmap created"):
        font.save(tempname)


def test_to_imagefont() -> None:
    font = FontFile.FontFile()
    with pytest.raises(ValueError, match="No bitmap created"):
        font.to_imagefont()



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import Image, IptcImagePlugin, TiffImagePlugin, TiffTags

from .helper import assert_image_equal

TEST_FILE = "Tests/images/iptc.jpg"


def create_iptc_image(info: dict[str, int] = {}) -> BytesIO:
    def field(tag: tuple[int, int], value: bytes) -> bytes:
        return bytes((0x1C,) + tag + (0, len(value))) + value

    data = field((3, 60), bytes((info.get("layers", 1), info.get("component", 0))))
    data += field((3, 120), bytes((info.get("compression", 1),)))
    if "band" in info:
        data += field((3, 65), bytes((info["band"] + 1,)))
    data += field((3, 20), b"\x01")  # width
    data += field((3, 30), b"\x01")  # height
    data += field(
        (8, 10),
        bytes((info.get("data", 0),)),
    )

    return BytesIO(data)


def test_open() -> None:
    expected = Image.new("L", (1, 1))

    f = create_iptc_image()
    with Image.open(f) as im:
        assert im.tile == [("iptc", (0, 0, 1, 1), 25, ("raw", None))]
        assert_image_equal(im, expected)

    with Image.open(f) as im:
        assert im.load() is not None


def test_field_length() -> None:
    f = create_iptc_image()
    f.seek(28)
    f.write(b"\xff")
    with pytest.raises(OSError, match="illegal field length in IPTC/NAA file"):
        with Image.open(f):
            pass


@pytest.mark.parametrize("layers, mode", ((3, "RGB"), (4, "CMYK")))
def test_layers(layers: int, mode: str) -> None:
    for band in range(-1, layers):
        info = {"layers": layers, "component": 1, "data": 5}
        if band != -1:
            info["band"] = band
        f = create_iptc_image(info)
        with Image.open(f) as im:
            assert im.mode == mode

            data = [0] * layers
            data[max(band, 0)] = 5
            assert im.getpixel((0, 0)) == tuple(data)


def test_unknown_compression() -> None:
    f = create_iptc_image({"compression": 2})
    with pytest.raises(OSError, match="Unknown IPTC image compression"):
        with Image.open(f):
            pass


def test_getiptcinfo() -> None:
    f = create_iptc_image()
    with Image.open(f) as im:
        assert IptcImagePlugin.getiptcinfo(im) == {
            (3, 60): b"\x01\x00",
            (3, 120): b"\x01",
            (3, 20): b"\x01",
            (3, 30): b"\x01",
        }


def test_getiptcinfo_jpg_none() -> None:
    # Arrange
    with Image.open("Tests/images/hopper.jpg") as im:
        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

    # Assert
    assert iptc is None


def test_getiptcinfo_jpg_found() -> None:
    # Arrange
    with Image.open(TEST_FILE) as im:
        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

    # Assert
    assert isinstance(iptc, dict)
    assert iptc[(2, 90)] == b"Budapest"
    assert iptc[(2, 101)] == b"Hungary"


def test_getiptcinfo_fotostation() -> None:
    # Arrange
    with open(TEST_FILE, "rb") as fp:
        data = bytearray(fp.read())
    data[86] = 240
    f = BytesIO(data)
    with Image.open(f) as im:
        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

    # Assert
    assert iptc is not None
    assert 240 in (tag[0] for tag in iptc.keys()), "FotoStation tag not found"


def test_getiptcinfo_zero_padding() -> None:
    # Arrange
    with Image.open(TEST_FILE) as im:
        im.info["photoshop"][0x0404] += b"\x00\x00\x00"

        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

    # Assert
    assert isinstance(iptc, dict)
    assert len(iptc) == 3


def test_getiptcinfo_tiff() -> None:
    expected = {(1, 90): b"\x1b%G", (2, 0): b"\xcf\xc0"}

    with Image.open("Tests/images/hopper.Lab.tif") as im:
        iptc = IptcImagePlugin.getiptcinfo(im)

    assert iptc == expected

    # Test with LONG tag type
    with Image.open("Tests/images/hopper.Lab.tif") as im:
        assert isinstance(im, TiffImagePlugin.TiffImageFile)
        im.tag_v2.tagtype[TiffImagePlugin.IPTC_NAA_CHUNK] = TiffTags.LONG
        iptc = IptcImagePlugin.getiptcinfo(im)

    assert iptc == expected


def test_getiptcinfo_tiff_none() -> None:
    # Arrange
    with Image.open("Tests/images/hopper.tif") as im:
        # Act
        iptc = IptcImagePlugin.getiptcinfo(im)

    # Assert
    assert iptc is None



from typing import Any

class ImagingCore:
    def __getitem__(self, index: int) -> float | tuple[int, ...] | None: ...
    def __getattr__(self, name: str) -> Any: ...

class ImagingFont:
    def __getattr__(self, name: str) -> Any: ...

class ImagingDraw:
    def __getattr__(self, name: str) -> Any: ...

class PixelAccess:
    def __getitem__(self, xy: tuple[int, int]) -> float | tuple[int, ...]: ...
    def __setitem__(
        self, xy: tuple[int, int], color: float | tuple[int, ...]
    ) -> None: ...

class ImagingDecoder:
    def __getattr__(self, name: str) -> Any: ...

class ImagingEncoder:
    def __getattr__(self, name: str) -> Any: ...

class _Outline:
    def close(self) -> None: ...
    def __getattr__(self, name: str) -> Any: ...

def font(image: ImagingCore, glyphdata: bytes) -> ImagingFont: ...
def outline() -> _Outline: ...
def __getattr__(name: str) -> Any: ...
